@@ -20,6 +20,11 @@
   http://github.com/jeromerobert/hmat-oss
 */
 
+#ifndef _GNU_SOURCE
+// for getline with old glibc
+#define _GNU_SOURCE
+#endif
+
 #include <stdio.h>
 #include <math.h>
 #ifdef __cplusplus
@@ -296,14 +301,13 @@ int main(int argc, char **argv) {
   hmat_cluster_tree_t* cluster_tree;
   hmat_matrix_t * hmatrix;
   int kLowerSymmetric = 1; /* =0 if not Symmetric */
-  int rc;
 
   problem_data_t problem_data;
   double l;
 
   int nrhs = 1;
-  double *drhs, *drhsCopy, derr;
-  float  *frhs, *frhsCopy, ferr;
+  double *drhs=NULL, *drhsCopy=NULL, derr;
+  float  *frhs=NULL, *frhsCopy=NULL, ferr;
 
   if (argc != 3) {
       fprintf(stderr, "Usage: %s pointsfilename (S|D)\n", argv[0]);
@@ -327,7 +331,6 @@ int main(int argc, char **argv) {
   hmat_get_parameters(&settings);
   hmat_init_default_interface(&hmat, type);
 
-  settings.compressionMethod = hmat_compress_aca_plus;
   /*settings->recompress = 0;*/
   /*settings->admissibilityFactor = 3.;*/
 
@@ -369,43 +372,49 @@ int main(int argc, char **argv) {
   cluster_tree = hmat_create_cluster_tree(points, 3, n, clustering);
   hmat_delete_clustering(clustering);
   printf("ClusterTree node count = %d\n", hmat_tree_nodes_count(cluster_tree));
-  hmatrix = hmat.create_empty_hmatrix(cluster_tree, cluster_tree, 0);
+  hmat_admissibility_t * admissibilityCondition = hmat_create_admissibility_standard(2.0);
+  hmatrix = hmat.create_empty_hmatrix_admissibility(
+            cluster_tree, cluster_tree, 0, admissibilityCondition);
+  hmat_delete_admissibility(admissibilityCondition);
   hmat.get_info(hmatrix, &mat_info);
   printf("HMatrix node count = %d\n", mat_info.nr_block_clusters);
 
   fprintf(stdout,"Assembly...");
-  rc = hmat.assemble_simple_interaction(hmatrix, &problem_data, interaction_real, kLowerSymmetric);
-  if (rc) {
-    fprintf(stderr, "Error in assembly, return code is %d, exiting...\n", rc);
-    hmat.finalize();
-    return rc;
-  }
+  hmat_assemble_context_t ctx_assemble;
+  hmat_assemble_context_init(&ctx_assemble);
+  ctx_assemble.user_context = &problem_data;
+  ctx_assemble.compression = hmat_create_compression_aca_plus(1e-5);
+  ctx_assemble.simple_compute = interaction_real;
+  ctx_assemble.lower_symmetric = kLowerSymmetric;
+  hmat.assemble_generic(hmatrix, &ctx_assemble);
   fprintf(stdout, "done.\n");
+  hmat_delete_compression(ctx_assemble.compression);
 
   hmat.get_info(hmatrix, &mat_info);
-  printf("Rk size = %ld\n", mat_info.compressed_size);
-  hmat_procedure_t* trunc = hmat_create_procedure_epsilon_truncate(type, 1.e-2);
+  printf("Rk size = %gM\n", 1e-6 * mat_info.compressed_size);
+  hmat.set_low_rank_epsilon(hmatrix, 1.e-2);
   fprintf(stdout,"Post-process Rk-matrices...");
-  hmat.walk(hmatrix, trunc);
-  hmat_delete_procedure(trunc);
+  hmat.truncate(hmatrix);
   fprintf(stdout, "done.\n");
   hmat.get_info(hmatrix, &mat_info);
-  printf("Rk size = %ld\n", mat_info.compressed_size);
+  printf("Rk size = %gM\n", 1e-6 * mat_info.compressed_size);
 
   fprintf(stdout,"Factorisation...");
-  rc = hmat.factorize(hmatrix, hmat_factorization_lu);
-  if (rc) {
-    fprintf(stderr, "Error in factorisation, return code is %d, exiting...\n", rc);
-    hmat.finalize();
-    return rc;
-  }
+  hmat_factorization_context_t ctx_facto;
+  hmat_factorization_context_init(&ctx_facto);
+  ctx_facto.factorization = hmat_factorization_lu;
+  hmat.factorize_generic(hmatrix, &ctx_facto);
   fprintf(stdout, "done.\n");
 
   fprintf(stdout,"Solve...");
   if(type == HMAT_SIMPLE_PRECISION){
-    hmat.solve_systems(hmatrix, frhs, nrhs);
+    hmat.vector_reorder(frhs, cluster_tree, 0, NULL, nrhs);
+    hmat.solve_dense(hmatrix, frhs, nrhs);
+    hmat.vector_restore(frhs, cluster_tree, 0, NULL, nrhs);
   }else{
-    hmat.solve_systems(hmatrix, drhs, nrhs);
+    hmat.vector_reorder(drhs, cluster_tree, 0, NULL, nrhs);
+    hmat.solve_dense(hmatrix, drhs, nrhs);
+    hmat.vector_restore(drhs, cluster_tree, 0, NULL, nrhs);
   }
   fprintf(stdout, "done.\n");
 
