@@ -22,10 +22,11 @@
 
 /** Sample application for the HMatrix library.
 */
-#include "full_matrix.hpp"
+#include "admissibility.hpp"
 #include "assembly.hpp"
 #include "data_types.hpp"
 #include "hmat_cpp_interface.hpp"
+#include "default_engine.hpp"
 #include <cstring>
 #include <iostream>
 #include <cmath>
@@ -53,27 +54,24 @@ distanceTo(const DofCoordinates& points, int i, const Point& to)
   return r;
 }
 
-class KrigingAssemblyFunction : public SimpleAssemblyFunction<D_t> {
-private:
-  const DofCoordinates& points;
-  double l;
+class KrigingFunction {
+    const DofCoordinates& points;
+    double l;
 
 public:
-  /** Constructor.
+    KrigingFunction(const DofCoordinates& _points, double _l):
+         points(_points) , l(_l) {}
 
-      \param _mat The FullMatrix<T> the values are taken from.
-   */
-  KrigingAssemblyFunction(const DofCoordinates& _points, double _l)
-    : SimpleAssemblyFunction<D_t>(), points(_points), l(_l) {}
+    D_t interaction(int i, int j) const {
+        return exp(-fabs(distanceTo(points, i, j)) / l);
+    }
 
-  D_t interaction(int i, int j) const {
-    // Exponential
-    return exp(-fabs(distanceTo(points, i, j)) / l);
-  }
+    // static for C API
+    static void compute(void* me, int i, int j, void* result) {
+        *static_cast<D_t*>(result) = static_cast<KrigingFunction *>(me)->interaction(i, j);
+    }
 };
 
-
-
 void readPointsFromFile(const char* filename, std::vector<Point>& points) {
   std::ifstream f(filename);
   std::string line;
@@ -84,9 +82,9 @@ void readPointsFromFile(const char* filename, std::vector<Point>& points) {
   }
 }
 
-FullMatrix<D_t>* createRhs(const DofCoordinates& coord, double l) {
+ScalarArray<D_t>* createRhs(const DofCoordinates& coord, double l) {
   const int n = (int) coord.size();
-  FullMatrix<D_t>* rhs = FullMatrix<D_t>::Zero(n, 1);
+  ScalarArray<D_t>* rhs = new ScalarArray<D_t>(n, 1);
 
   Point center(0., 0., 0.);
   for (int i = 0; i < n; i++) {
@@ -123,10 +121,9 @@ double correlationLength(const std::vector<Point>& points) {
 
 template<template<typename> class E>
 int go(const char* pointsFilename) {
-  if (0 != HMatInterface<D_t, E>::init()) return 1;
+  if (0 != E<D_t>::init()) return 1;
 
   HMatSettings& settings = HMatSettings::getInstance();
-  settings.compressionMethod = AcaPlus;
   settings.setParameters();
 
   std::cout << "Load points...";
@@ -144,17 +141,20 @@ int go(const char* pointsFilename) {
   DofCoordinates coord(xyz, 3, points.size(), true);
 
   const double l = correlationLength(points);
-  FullMatrix<D_t>* rhs = createRhs(coord, l);
-  FullMatrix<D_t> rhsCopy(n, 1);
+  ScalarArray<D_t>* rhs = createRhs(coord, l);
+  ScalarArray<D_t> rhsCopy(n, 1);
   rhsCopy.copyMatrixAtOffset(rhs, 0, 0);
-
-  KrigingAssemblyFunction f(coord, l);
+  KrigingFunction kriginFunction(coord, l);
+  AssemblyFunction<D_t, SimpleFunction> f(
+      SimpleFunction<D_t>(&KrigingFunction::compute, &kriginFunction), new CompressionAcaPlus(1e-4));
 
   ClusterTree* ct = createClusterTree(coord);
   std::cout << "ClusterTree node count = " << ct->nodesCount() << std::endl;
   // Either store lower triangular matrix and use LDLt (or LLt) factorization or
   // store full matrix and use LU factorization.
-  HMatInterface<D_t, E> hmat(ct, ct, kNotSymmetric);
+  IEngine<D_t>* engine = new E<D_t>();
+  StandardAdmissibilityCondition * admissibility = new StandardAdmissibilityCondition(2.0);
+  HMatInterface<D_t> hmat(engine, ct, ct, kNotSymmetric, admissibility);
   settings.setParameters();
 
   hmat.assemble(f, kLowerSymmetric);
@@ -166,7 +166,7 @@ int go(const char* pointsFilename) {
 
   std::cout << "done.\nFactorisation...";
 
-  hmat.factorize(hmat_factorization_lu);
+  hmat.factorize(hmat::Factorization::LU);
 
   std::cout << "Resolution...";
   hmat.solve(*rhs);
@@ -176,7 +176,7 @@ int go(const char* pointsFilename) {
   double rhsCopyNorm = rhsCopy.norm();
   for (int i = 0; i < n; i++) {
     for (int j = 0; j < n; j++) {
-      rhsCopy.get(i, 0) -= f.interaction(i, j) * rhs->get(j, 0);
+      rhsCopy.get(i, 0) -= kriginFunction.interaction(i, j) * rhs->get(j, 0);
     }
   }
   double diffNorm = rhsCopy.norm();
@@ -184,7 +184,8 @@ int go(const char* pointsFilename) {
   std::cout << "||Ax - b|| / ||b|| = " << diffNorm / rhsCopyNorm << std::endl;
 
   delete rhs;
-  HMatInterface<D_t, E>::finalize();
+  delete admissibility;
+  E<D_t>::finalize();
   return 0;
 }
 
