@@ -26,6 +26,7 @@
 
 #include "config.h"
 
+#include <assert.h>
 #include "data_types.hpp"
 
 #ifdef HAVE_MKL_CBLAS_H
@@ -74,30 +75,6 @@ void axpy(const int n, const hmat::Z_t& alpha, const hmat::Z_t* x, const int inc
   #undef _C_T
 }
 
-template<typename T>
-void copy(const int n, const T* x, const int incx, T* y, const int incy);
-
-inline
-void copy(const int n, const hmat::S_t* x, int incx, hmat::S_t* y, const int incy) {
-  cblas_scopy(n, x, incx, y, incy);
-}
-inline
-void copy(const int n, const hmat::D_t* x, int incx, hmat::D_t* y, const int incy) {
-  cblas_dcopy(n, x, incx, y, incy);
-}
-inline
-void copy(const int n, const hmat::C_t* x, int incx, hmat::C_t* y, const int incy) {
-  #define _C_T hmat::C_t
-  cblas_ccopy(n, _C(x), incx, _C(y), incy);
-  #undef _C_T
-}
-inline
-void copy(const int n, const hmat::Z_t* x, int incx, hmat::Z_t* y, const int incy) {
-  #define _C_T hmat::Z_t
-  cblas_zcopy(n, _C(x), incx, _C(y), incy);
-  #undef _C_T
-}
-
 template<typename T>
 T dot(const int n, const T* x, const int incx, const T* y, const int incy);
 
@@ -216,19 +193,19 @@ void gemv(const char trans, const int m, const int n, const T& alpha, const T* a
 inline
 void gemv(const char trans, const int m, const int n, const hmat::S_t& alpha, const hmat::S_t* a, const int lda,
           const hmat::S_t* x, const int incx, const hmat::S_t& beta, hmat::S_t* y, const int incy) {
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   cblas_sgemv(CblasColMajor, t, m, n, alpha, a, lda, x, incx, beta, y, incy);
 }
 inline
 void gemv(const char trans, const int m, const int n, const hmat::D_t& alpha, const hmat::D_t* a, const int lda,
           const hmat::D_t* x, const int incx, const hmat::D_t& beta, hmat::D_t* y, const int incy) {
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   cblas_dgemv(CblasColMajor, t, m, n, alpha, a, lda, x, incx, beta, y, incy);
 }
 inline
 void gemv(const char trans, const int m, const int n, const hmat::C_t& alpha, const hmat::C_t* a, const int lda,
           const hmat::C_t* x, const int incx, const hmat::C_t& beta, hmat::C_t* y, const int incy) {
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   // WARNING: &alpha/&beta instead of alpha/beta for complex values
   #define _C_T hmat::C_t
   cblas_cgemv(CblasColMajor, t, m, n, _C(&alpha), _C(a), lda, _C(x), incx, _C(&beta), _C(y), incy);
@@ -237,7 +214,7 @@ void gemv(const char trans, const int m, const int n, const hmat::C_t& alpha, co
 inline
 void gemv(const char trans, const int m, const int n, const hmat::Z_t& alpha, const hmat::Z_t* a, const int lda,
           const hmat::Z_t* x, const int incx, const hmat::Z_t& beta, hmat::Z_t* y, const int incy) {
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   // WARNING: &alpha/&beta instead of alpha/beta for complex values
   #define _C_T hmat::Z_t
   cblas_zgemv(CblasColMajor, t, m, n, _C(&alpha), _C(a), lda, _C(x), incx, _C(&beta), _C(y), incy);
@@ -285,46 +262,50 @@ inline
 void gemm(const char transA, const char transB, const int m, const int n, const int k,
           const hmat::S_t& alpha, const hmat::S_t* a, const int lda, const hmat::S_t* b, const int ldb,
           const hmat::S_t& beta, hmat::S_t* c, const int ldc) {
-  const CBLAS_TRANSPOSE tA = (transA == 'N' ? CblasNoTrans : CblasTrans);
-  const CBLAS_TRANSPOSE tB = (transB == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE tA = (transA == 'C' ? CblasConjTrans : (transA == 'T' ? CblasTrans : CblasNoTrans));
+  const CBLAS_TRANSPOSE tB = (transB == 'C' ? CblasConjTrans : (transB == 'T' ? CblasTrans : CblasNoTrans));
   cblas_sgemm(CblasColMajor, tA, tB, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc);
 }
 inline
 void gemm(const char transA, const char transB, const int m, const int n, const int k,
           const hmat::D_t& alpha, const hmat::D_t* a, const int lda, const hmat::D_t* b, const int ldb,
           const hmat::D_t& beta, hmat::D_t* c, const int ldc) {
-  const CBLAS_TRANSPOSE tA = (transA == 'N' ? CblasNoTrans : CblasTrans);
-  const CBLAS_TRANSPOSE tB = (transB == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE tA = (transA == 'C' ? CblasConjTrans : (transA == 'T' ? CblasTrans : CblasNoTrans));
+  const CBLAS_TRANSPOSE tB = (transB == 'C' ? CblasConjTrans : (transB == 'T' ? CblasTrans : CblasNoTrans));
   cblas_dgemm(CblasColMajor, tA, tB, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc);
 }
 inline
 void gemm(const char transA, const char transB, const int m, const int n, const int k,
           const hmat::C_t& alpha, const hmat::C_t* a, const int lda, const hmat::C_t* b, const int ldb,
           const hmat::C_t& beta, hmat::C_t* c, const int ldc) {
-  const CBLAS_TRANSPOSE tA = (transA == 'N' ? CblasNoTrans : CblasTrans);
-  const CBLAS_TRANSPOSE tB = (transB == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE tA = (transA == 'C' ? CblasConjTrans : (transA == 'T' ? CblasTrans : CblasNoTrans));
+  const CBLAS_TRANSPOSE tB = (transB == 'C' ? CblasConjTrans : (transB == 'T' ? CblasTrans : CblasNoTrans));
   // WARNING: &alpha/&beta instead of alpha/beta for complex values
 #define _C_T hmat::C_t
 #ifdef HAVE_ZGEMM3M
-  cblas_cgemm3m(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
-#else
-  cblas_cgemm(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
+  static char* noGemm3m=getenv("HMAT_NO_GEMM3M"); // forbids inlining :-(
+  if (!noGemm3m)
+    cblas_cgemm3m(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
+  else
 #endif
+    cblas_cgemm(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
 #undef _C_T
 }
 inline
 void gemm(const char transA, const char transB, const int m, const int n, int k,
           const hmat::Z_t& alpha, const hmat::Z_t* a, const int lda, const hmat::Z_t* b, const int ldb,
           const hmat::Z_t& beta, hmat::Z_t* c, const int ldc) {
-  const CBLAS_TRANSPOSE tA = (transA == 'N' ? CblasNoTrans : CblasTrans);
-  const CBLAS_TRANSPOSE tB = (transB == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE tA = (transA == 'C' ? CblasConjTrans : (transA == 'T' ? CblasTrans : CblasNoTrans));
+  const CBLAS_TRANSPOSE tB = (transB == 'C' ? CblasConjTrans : (transB == 'T' ? CblasTrans : CblasNoTrans));
   // WARNING: &alpha/&beta instead of alpha/beta for complex values
 #define _C_T hmat::Z_t
 #ifdef HAVE_ZGEMM3M
-  cblas_zgemm3m(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
-#else
-  cblas_zgemm(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
+  static char* noGemm3m=getenv("HMAT_NO_GEMM3M"); // forbids inlining :-(
+  if (!noGemm3m)
+    cblas_zgemm3m(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
+  else
 #endif
+    cblas_zgemm(CblasColMajor, tA, tB, m, n, k, _C(&alpha), _C(a), lda, _C(b), ldb, _C(&beta), _C(c), ldc);
 #undef _C_T
 }
 
@@ -339,7 +320,7 @@ void trmm(const char side, const char uplo, const char trans, const char diag,
           hmat::S_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   cblas_strmm(CblasColMajor, s, u, t, d, m, n, alpha, a, lda, b, ldb);
 }
@@ -349,7 +330,7 @@ void trmm(const char side, const char uplo, const char trans, const char diag,
           hmat::D_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   cblas_dtrmm(CblasColMajor, s, u, t, d, m, n, alpha, a, lda, b, ldb);
 }
@@ -359,7 +340,7 @@ void trmm(const char side, const char uplo, const char trans, const char diag,
           hmat::C_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   // WARNING: &alpha instead of alpha for complex values
 #define _C_T hmat::C_t
@@ -372,7 +353,7 @@ void trmm(const char side, const char uplo, const char trans, const char diag,
           hmat::Z_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   // WARNING: &alpha instead of alpha for complex values
 #define _C_T hmat::Z_t
@@ -392,7 +373,7 @@ void trsm(const char side, const char uplo, const char trans, const char diag,
   assert(lda >= m || side != 'L');
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   cblas_strsm(CblasColMajor, s, u, t, d, m, n, alpha, a, lda, b, ldb);
 }
@@ -402,7 +383,7 @@ void trsm(const char side, const char uplo, const char trans, const char diag,
           hmat::D_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   cblas_dtrsm(CblasColMajor, s, u, t, d, m, n, alpha, a, lda, b, ldb);
 }
@@ -412,7 +393,7 @@ void trsm(const char side, const char uplo, const char trans, const char diag,
           hmat::C_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   // WARNING: &alpha instead of alpha for complex values
 #define _C_T hmat::C_t
@@ -425,7 +406,7 @@ void trsm(const char side, const char uplo, const char trans, const char diag,
           hmat::Z_t* b, const int ldb) {
   const CBLAS_SIDE s = (side == 'L' ? CblasLeft : CblasRight);
   const CBLAS_UPLO u = (uplo == 'U' ? CblasUpper : CblasLower);
-  const CBLAS_TRANSPOSE t = (trans == 'N' ? CblasNoTrans : CblasTrans);
+  const CBLAS_TRANSPOSE t = (trans == 'C' ? CblasConjTrans : (trans == 'T' ? CblasTrans : CblasNoTrans));
   const CBLAS_DIAG d = (diag == 'N' ?  CblasNonUnit : CblasUnit );
   // WARNING: &alpha instead of alpha for complex values
 #define _C_T hmat::Z_t
