@@ -84,7 +84,7 @@ public:
     */
   std::string description() const {
     std::ostringstream convert;   // stream used for the conversion
-    convert << "[" << offset() << ", " << offset()+size()-1 << "]" ;
+    convert << "[" << offset() << ", " << size() << "]" ;
     return convert.str();
   }
 
@@ -105,7 +105,7 @@ public:
   ~DofData();
 
   DofData* copy() const;
-  inline int size() const { return coordinates_->size(); }
+  inline int size() const { return coordinates_->numberOfDof(); }
 
 private:
   /// Indices array
@@ -122,11 +122,22 @@ public:
   ClusterData(const DofData* dofData) : IndexSet(0, dofData->size()), dofData_(dofData) {}
   ClusterData(const ClusterData& data) : IndexSet(data), dofData_(data.dofData_) {}
   ClusterData(const ClusterData& data, int offset, int size) : IndexSet(offset, size), dofData_(data.dofData_) {}
-
+  /**
+   * @brief indices()[i] is the index in the original coordinate
+   * array which match the ith point in the underlying hmat/dofData_
+   */
   inline int* indices() const { return dofData_->perm_i2e_; }
+  /**
+   * @brief indices_rev()[i] is the index in the underlying hmat/dofData_
+   * which match the ith point in the original coordinate array
+   */
   inline int* indices_rev() const { return dofData_->perm_e2i_; }
   inline const DofCoordinates* coordinates() const { return dofData_->coordinates_; }
   inline int* group_index() const { return dofData_->group_index_; }
+  /*! Move a degree of freedom into its right sibling  */
+  void moveDoF(int index, ClusterData* right);
+  /** @brief debug method that check the underlying indices arrays are valid */
+  void assertValid();
 
 private:
   const DofData* dofData_;
@@ -147,10 +158,8 @@ public:
   /*! Data */
   ClusterData data;
 
-  /*! Opaque pointer which may be used by clustering algorithm */
-  mutable void* clusteringAlgoData_;
-  /*! Opaque pointer which may be used by admissibility algorithm */
-  mutable void* admissibilityAlgoData_;
+  /*! Opaque pointer which may be used by algorithms */
+  mutable void* cache_;
 
 public:
   /*! \brief Create a leaf.
@@ -160,38 +169,19 @@ public:
       \param _threshold max number of indices in a leaf. Used for the recursive division.
    */
   explicit ClusterTree(const DofData* dofData);
+    explicit ClusterTree(const DofData* dofData, int offset, int size);
 
   /* Copy constructor */
   ClusterTree(const ClusterTree& other);
 
   virtual ~ClusterTree();
 
-  /*! \brief Returns the admissibility parameter eta corresponding to two clusters.
-
-    As described in \a ClusterTree::isAdmissibleWith documentation,
-    this criteria is defined by:
-
-      eta = min(diameter(), other->diameter()) / distanceTo(other);
-
-    The lower it is, the farther away are the two boxes. It is thus
-    linked with the compression ratio one can expect from a block, the
-    lower eta is, the more the block can be compressed. It can be used
-    as a parameter in a crude a-priori work estimation method.
-   */
-  double getEta(const ClusterTree* other) const;
-
   ClusterTree* slice(int offset, int size) const;
 
   /*! \brief Return a copy to this.
    */
   ClusterTree* copy(const ClusterTree* copyFather=NULL) const;
 
-  /**
-   * Transform this cluster tree so it has the same or greater depth
-   * as the other one.
-   */
-  void sameDepth(const ClusterTree * other);
-
   /*! \brief Return a short string describing the content of this ClusterTree for debug (like: "[320, 452]")
     */
   std::string description() const {
@@ -202,17 +192,21 @@ public:
 
 class AxisAlignedBoundingBox
 {
+    const unsigned dimension_;
+    double * bb_;
 public:
-  explicit AxisAlignedBoundingBox(const ClusterData& node);
-  AxisAlignedBoundingBox(int dim, const double *bboxMin, const double *bboxMax);
-  ~AxisAlignedBoundingBox();
-
-  double diameter() const;
-  double distanceTo(const AxisAlignedBoundingBox& other) const;
-private:
-  const int dimension_;
-public:
-  double *bbMin, *bbMax;
+    explicit AxisAlignedBoundingBox(const ClusterData& node);
+    ~AxisAlignedBoundingBox();
+    double extends(int dim) const;
+    int greatestDim() const;
+    double diameter() const;
+    double diameterSqr() const;
+    double distanceTo(const AxisAlignedBoundingBox& other) const;
+    double distanceToSqr(const AxisAlignedBoundingBox& other) const;
+    const double * bbMin() const { return bb_; }
+    const double * bbMax() const { return bb_ + dimension_; }
+    void bbMin(const double * bbMin);
+    void bbMax(const double * bbMax);
 };
 
 }  // end namespace hmat
