@@ -41,6 +41,8 @@
 #include "full_matrix.hpp"
 #include "common/context.hpp"
 #include "common/my_assert.h"
+#include "cluster_assembly_function.hpp"
+#include "random_pivot_manager.hpp"
 
 #ifdef _MSC_VER
 // Intel compiler defines isnan in global namespace
@@ -60,78 +62,6 @@ using std::max;
 
 namespace hmat {
 
-
-/* Convenience class to lighten the getRow() / getCol() / assemble calls
-   and use information from block_info_t to speed up things for sparse and null blocks.
-*/
-template<typename T>
-class ClusterAssemblyFunction {
-  const Function<T>& f;
-
-public:
-  const ClusterData* rows;
-  const ClusterData* cols;
-  hmat_block_info_t info;
-  const AllocationObserver & allocationObserver_;
-  ClusterAssemblyFunction(const Function<T>& _f,
-                          const ClusterData* _rows, const ClusterData* _cols,
-                          const AllocationObserver & allocationObserver)
-    : f(_f), rows(_rows), cols(_cols), allocationObserver_(allocationObserver) {
-    f.prepareBlock(rows, cols, &info, allocationObserver_);
-    assert((info.user_data == NULL) == (info.release_user_data == NULL));
-  }
-  ~ClusterAssemblyFunction() {
-    f.releaseBlock(&info, allocationObserver_);
-  }
-  void getRow(int index, Vector<typename Types<T>::dp>& result) const {
-    if (info.block_type != hmat_block_sparse || !info.is_null_row(&info, index))
-      f.getRow(rows, cols, index, info.user_data, &result);
-  }
-  void getCol(int index, Vector<typename Types<T>::dp>& result) const {
-    if (info.block_type != hmat_block_sparse || !info.is_null_col(&info, index))
-      f.getCol(rows, cols, index, info.user_data, &result);
-  }
-  FullMatrix<typename Types<T>::dp>* assemble() const {
-    if (info.block_type != hmat_block_null)
-      return f.assemble(rows, cols, &info, allocationObserver_) ;
-    else
-      // TODO return
-      return FullMatrix<typename Types<T>::dp>::Zero(rows->size(), cols->size());
-  }
-private:
-  ClusterAssemblyFunction(ClusterAssemblyFunction&o) {} // No copy
-};
-
-template<typename T> double squaredNorm(const T x) {
-  return x * x;
-}
-// Specializations for complex values
-template<> double squaredNorm(const C_t x) {
-// std::norm seems deadfully slow on Intel 15
-#ifdef __INTEL_COMPILER
-  const float x_r = x.real();
-  const float x_i = x.imag();
-  return x_r*x_r + x_i*x_i;
-#else
-  return std::norm(x);
-#endif
-}
-template<> double squaredNorm(const Z_t x) {
-#ifdef __INTEL_COMPILER
-  const double x_r = x.real();
-  const double x_i = x.imag();
-  return x_r*x_r + x_i*x_i;
-#else
-  return std::norm(x);
-#endif
-}
-
-template<typename T> static bool isZero(const Vector<T>& v) {
-  int index = v.absoluteMaxIndex();
-  return v.v[index] == Constants<T>::zero;
-}
-
-
 /** \brief Updates a row to reflect its current value in the matrix.
 
     \param rowVec the row to update
@@ -144,7 +74,7 @@ template<typename T>
 static void updateRow(Vector<T>& rowVec, int row, const vector<Vector<T>*>& rows,
                       const vector<Vector<T>*>& cols, int k) {
   for (int l = 0; l < k; l++) {
-    rowVec.axpy(Constants<T>::mone * cols[l]->v[row], rows[l]);
+    rowVec.axpy(Constants<T>::mone * (*cols[l])[row], rows[l]);
   }
 }
 
@@ -155,14 +85,18 @@ static void updateCol(Vector<T>& colVec, int col, const vector<Vector<T>*>& cols
 }
 
 
-template<typename T> static void findMax(FullMatrix<T>* m, int& i, int& j) {
-  i = 0;
-  j = 0;
-  double maxNorm = squaredNorm<T>(m->get(i, j));
-  const int cols = m->cols;
-  const int rows = m->rows;
-  for (int col = 0; col < cols; col++) {
-    for (int row = 0; row < rows; row++) {
+template<typename T> static void findMax(const FullMatrix<T>* m, int& i, int& j) {
+  if (m->data.lda == m->data.rows) {
+    // quick path
+    int k = proxy_cblas::i_amax(m->data.rows*m->data.cols, m->data.const_ptr(), 1);
+    i = k % m->data.rows ;
+    j = (k - i) / m->data.rows ;
+  } else {
+    i = 0;
+    j = 0;
+    double maxNorm = 0.;
+    for (int col = 0; col < m->cols(); col++) {
+      int row = proxy_cblas::i_amax(m->data.rows, m->data.const_ptr(0,col), 1);
       const double norm = squaredNorm<T>(m->get(row, col));
       if (norm > maxNorm) {
         i = row;
@@ -193,7 +127,7 @@ static int findCol(const ClusterAssemblyFunction<T>& block, vector<bool>& colFre
       col.clear();
       block.getCol(i, col);
       colFree[i] = false;
-      if (!isZero(col)) {
+      if (!col.isZero()) {
         found = true;
         break;
       }
@@ -221,7 +155,7 @@ static int findMinRow(const ClusterAssemblyFunction<T>& block,
     minNorm2 = DBL_MAX;
     for (int i = 0; i < rowCount; i++) {
       if (rowFree[i]) {
-        double norm2 = squaredNorm<typename Types<T>::dp>(aRef.v[i]);
+        double norm2 = squaredNorm<typename Types<T>::dp>(aRef[i]);
         if (norm2 < minNorm2) {
           i_ref = i;
           minNorm2 = norm2;
@@ -234,7 +168,7 @@ static int findMinRow(const ClusterAssemblyFunction<T>& block,
     row.clear();
     block.getRow(i_ref, row);
     updateRow<typename Types<T>::dp>(row, i_ref, bCols, aCols, aCols.size());
-    found = !isZero(row);
+    found = !row.isZero();
     rowFree[i_ref] = false;
   }
   return i_ref;
@@ -257,7 +191,7 @@ static int findMinCol(const ClusterAssemblyFunction<T>& block,
     minNorm2 = DBL_MAX;
     for (int j = 0; j < colCount; j++) {
       if (colFree[j]) {
-        double norm2 = squaredNorm<typename Types<T>::dp>(bRef.v[j]);
+        double norm2 = squaredNorm<typename Types<T>::dp>(bRef[j]);
         if (norm2 < minNorm2) {
           j_ref = j;
           minNorm2 = norm2;
@@ -270,7 +204,7 @@ static int findMinCol(const ClusterAssemblyFunction<T>& block,
     col.clear();
     block.getCol(j_ref, col);
     updateCol<typename Types<T>::dp>(col, j_ref, aCols, bCols, bCols.size());
-    found = !isZero(col);
+    found = !col.isZero();
     colFree[j_ref] = false;
   }
   return j_ref;
@@ -278,92 +212,66 @@ static int findMinCol(const ClusterAssemblyFunction<T>& block,
 
 
 template<typename T>
-RkMatrix<T>* compressMatrix(FullMatrix<T>* m, const IndexSet* rows,
-                            const IndexSet* cols) {
+RkMatrix<T>* truncatedSvd(FullMatrix<T>* m, double epsilon) {
   DECLARE_CONTEXT;
-  assert(m->rows == rows->size());
-  assert(m->cols == cols->size());
-  assert(m->lda >= m->rows);
 
-  //TODO replace with a case with m==NULL
-  bool zeroMatrix = true;
-  for (int col = 0; col < m->cols; col++) {
-    Vector<T> v(m->m + ((size_t) m->rows) * col, m->rows);
-    zeroMatrix = zeroMatrix && isZero(v);
-    if (!zeroMatrix) break;
+  if (m->isZero()) {
+    return new RkMatrix<T>(NULL, m->rows_, NULL, m->cols_);
   }
-  if (zeroMatrix) {
-    return new RkMatrix<T>(NULL, rows, NULL, cols, NoCompression);
-  }
-  // In the case of non-square matrix, we don't calculated singular vectors
+  // In the case of non-square matrix, we don't calculate singular vectors
   // bigger than the minimum dimension of the matrix. However this is not
   // necessary here, since k < min (n, p) for M matrix (nxp).
-  int rowCount = m->rows;
-  int colCount = m->cols;
-  FullMatrix<T> *u = NULL, *vt = NULL;
-  Vector<double>* sigma = NULL;
+  ScalarArray<T> *u = NULL, *v = NULL;
 
-  int info = truncatedSvd<T>(m, &u, &sigma, &vt);
-  HMAT_ASSERT(info == 0);
-  // Control of the approximation
+  // TODO compress with something else than SVD
+  m->data.truncatedSvdDecomposition(&u, &v, epsilon);
 
-  int maxK = min(rowCount, colCount);
-  int k = RkMatrix<T>::approx.findK(sigma->v, maxK, RkMatrix<T>::approx.assemblyEpsilon);
-
-  if(k == 0)
-    return new RkMatrix<T>(NULL, rows, NULL, cols, NoCompression);
-
-  for (int col = 0; col < k; col++) {
-    for (int row = 0; row < rowCount; row++) {
-      u->get(row, col) *= sigma->v[col];
-    }
-  }
-
-  FullMatrix<T> matU(u->m, rowCount, k);
-  FullMatrix<T>* uTilde = matU.copy();
-  FullMatrix<T> matV(vt->m, k, colCount, maxK);
-  FullMatrix<T>* vTilde = matV.copyAndTranspose();
-
-  delete u;
-  delete vt;
-  delete sigma;
-  FullMatrix<T>* a = uTilde;
-  return new RkMatrix<T>(a, rows, vTilde, cols, Svd);
+  return new RkMatrix<T>(u, m->rows_, v, m->cols_);
 }
 
 
 template<typename T>
-static RkMatrix<typename Types<T>::dp>*
-compressSvd(const ClusterAssemblyFunction<T>& block) {
+RkMatrix<typename Types<T>::dp>*
+doCompressionSVD(const ClusterAssemblyFunction<T>& block, double compressionEpsilon) {
   DECLARE_CONTEXT;
   typedef typename Types<T>::dp dp_t;
-  // TODO: use ClusterAssemblyFunction to optimize with blockinfo_t
   FullMatrix<dp_t>* m = block.assemble();
-  RkMatrix<dp_t>* result = compressMatrix(m, block.rows, block.cols);
+  RkMatrix<dp_t>* result = truncatedSvd(m, compressionEpsilon);
   delete m;
   return result;
 }
 
+RkMatrix<Types<S_t>::dp>*
+CompressionSVD::compress(const ClusterAssemblyFunction<S_t>& block) const {
+    return doCompressionSVD<S_t>(block, epsilon_);
+}
+RkMatrix<Types<D_t>::dp>*
+CompressionSVD::compress(const ClusterAssemblyFunction<D_t>& block) const {
+    return doCompressionSVD<D_t>(block, epsilon_);
+}
+RkMatrix<Types<C_t>::dp>*
+CompressionSVD::compress(const ClusterAssemblyFunction<C_t>& block) const {
+    return doCompressionSVD<C_t>(block, epsilon_);
+}
+RkMatrix<Types<Z_t>::dp>*
+CompressionSVD::compress(const ClusterAssemblyFunction<Z_t>& block) const {
+    return doCompressionSVD<Z_t>(block, epsilon_);
+}
+
 
 template<typename T>
-static RkMatrix<typename Types<T>::dp>*
-compressAcaFull(const ClusterAssemblyFunction<T>& block) {
+RkMatrix<typename Types<T>::dp>*
+doCompressionAcaFull(const ClusterAssemblyFunction<T>& block, double compressionEpsilon) {
   DECLARE_CONTEXT;
   typedef typename Types<T>::dp dp_t;
   // TODO: use ClusterAssemblyFunction to optimize with blockinfo_t
   FullMatrix<dp_t>* m = block.assemble();
 
-  const double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
   double estimateSquaredNorm = 0;
-  int maxK = min(m->rows, m->cols);
-  if (RkMatrix<dp_t>::approx.k > 0) {
-    maxK = min(maxK, RkMatrix<dp_t>::approx.k);
-  }
+  int maxK = min(m->rows(), m->cols());
 
-  FullMatrix<dp_t> tmpA(m->rows, maxK);
-  tmpA.clear();
-  FullMatrix<dp_t> tmpB(m->cols, maxK);
-  tmpB.clear();
+  ScalarArray<dp_t> tmpA(m->rows(), maxK);
+  ScalarArray<dp_t> tmpB(m->cols(), maxK);
   int nu;
 
   for (nu = 0; nu < maxK; nu++) {
@@ -375,28 +283,28 @@ compressAcaFull(const ClusterAssemblyFunction<T>& block) {
     }
 
     // Creation of the vectors A_i_nu and B_j_nu
-    memcpy(tmpA.m + nu * tmpA.rows, m->m + j_nu * m->rows,
-           sizeof(dp_t) * tmpA.rows);
-    for (int j = 0; j < m->cols; j++) {
-      tmpB.get(j, nu) = m->get(i_nu, j) / delta;
-    }
-
-    proxy_cblas::ger(m->rows, m->cols, Constants<dp_t>::mone, tmpA.m + nu * tmpA.rows, 1, tmpB.m + nu * tmpB.rows, 1, m->m, m->rows);
-
-    // Update the estimate norm
-    // Let S_{k-1} be the previous estimate. We have (for the Frobenius norm):
-    //  ||S_k||^2 = ||S_{k-1}||^2 + \sum_{l = 0}^{nu-1} (<a_k, a_l> <b_k, b_l> + <a_l, a_k> <b_l, b_k>))
-    //              + ||a_k||^2 ||b_k||^2
     {
-      Vector<dp_t> va_nu(tmpA.m + nu * tmpA.rows, tmpA.rows);
-      Vector<dp_t> vb_nu(tmpB.m + nu * tmpB.rows, tmpB.rows);
-      Vector<dp_t> a_l(tmpA.m, tmpA.rows);
-      Vector<dp_t> b_l(tmpB.m, tmpB.rows);
+      Vector<dp_t> va_nu(tmpA, nu);
+      Vector<dp_t> vb_nu(tmpB, nu);
+
+      for (int i = 0; i < m->rows(); i++)
+        va_nu[i] = m->get(i, j_nu);
+      for (int j = 0; j < m->cols(); j++)
+        vb_nu[j] = m->get(i_nu, j) / delta;
+
+      // performs the rank 1 operation m := m - va_nu*vb_nu^T
+      // in order to nullify m->get(i_nu, j_nu) (the previous maximum value)
+      m->data.rankOneUpdate(Constants<dp_t>::mone, va_nu, vb_nu);
+
+      // Update the estimate norm
+      // Let S_{k-1} be the previous estimate. We have (for the Frobenius norm):
+      //  ||S_k||^2 = ||S_{k-1}||^2 + \sum_{l = 0}^{nu-1} (<a_k, a_l> <b_k, b_l> + <a_l, a_k> <b_l, b_k>))
+      //              + ||a_k||^2 ||b_k||^2
       // The sum
       double newEstimate = 0.0;
       for (int l = 0; l < nu - 1; l++) {
-        a_l.v = tmpA.m + l * tmpA.rows;
-        b_l.v = tmpB.m + l * tmpB.rows;
+        Vector<dp_t> a_l(tmpA, l);
+        Vector<dp_t> b_l(tmpB, l);
         newEstimate += hmat::real(Vector<dp_t>::dot(&va_nu, &a_l) * Vector<dp_t>::dot(&vb_nu, &b_l));
       }
       estimateSquaredNorm += 2.0 * newEstimate;
@@ -406,9 +314,9 @@ compressAcaFull(const ClusterAssemblyFunction<T>& block) {
       estimateSquaredNorm += ab_norm_2;
 
       // Evaluate the stopping criterion
-      // ||a_nu|| ||b_nu|| < epsilon * ||S_nu||
-      // <=> ||a_nu||^2 ||b_nu||^2 < epsilon^2 ||S_nu||^2
-      if (ab_norm_2 < epsilon * epsilon * estimateSquaredNorm) {
+      // ||a_nu|| ||b_nu|| < compressionEpsilon * ||S_nu||
+      // <=> ||a_nu||^2 ||b_nu||^2 < compressionEpsilon^2 ||S_nu||^2
+      if (ab_norm_2 < compressionEpsilon * compressionEpsilon * estimateSquaredNorm) {
         break;
       }
     }
@@ -416,26 +324,37 @@ compressAcaFull(const ClusterAssemblyFunction<T>& block) {
   delete m;
 
   if (nu == 0) {
-    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols, AcaFull);
+    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols);
   }
 
-  FullMatrix<dp_t>* newA = new FullMatrix<dp_t>(tmpA.rows, nu);
-  newA->clear();
-  memcpy(newA->m, tmpA.m, sizeof(dp_t) * tmpA.rows * nu);
-  FullMatrix<dp_t>* newB = new FullMatrix<dp_t>(tmpB.rows, nu);
-  newB->clear();
-  memcpy(newB->m, tmpB.m, sizeof(dp_t) * tmpB.rows * nu);
+  tmpA.cols=nu; // resize tmpA and tmpB for the copy
+  tmpB.cols=nu;
 
-  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols, AcaFull);
+  return new RkMatrix<dp_t>(tmpA.copy(), block.rows, tmpB.copy(), block.cols);
 }
 
+RkMatrix<Types<S_t>::dp>*
+CompressionAcaFull::compress(const ClusterAssemblyFunction<S_t>& block) const {
+    return doCompressionAcaFull<S_t>(block, epsilon_);
+}
+RkMatrix<Types<D_t>::dp>*
+CompressionAcaFull::compress(const ClusterAssemblyFunction<D_t>& block) const {
+    return doCompressionAcaFull<D_t>(block, epsilon_);
+}
+RkMatrix<Types<C_t>::dp>*
+CompressionAcaFull::compress(const ClusterAssemblyFunction<C_t>& block) const {
+    return doCompressionAcaFull<C_t>(block, epsilon_);
+}
+RkMatrix<Types<Z_t>::dp>*
+CompressionAcaFull::compress(const ClusterAssemblyFunction<Z_t>& block) const {
+    return doCompressionAcaFull<Z_t>(block, epsilon_);
+}
 
 template<typename T>
-static RkMatrix<typename Types<T>::dp>*
-compressAcaPartial(const ClusterAssemblyFunction<T>& block) {
+RkMatrix<typename Types<T>::dp>*
+doCompressionAcaPartial(const ClusterAssemblyFunction<T>& block, double compressionEpsilon, bool useRandomPivots) {
   typedef typename Types<T>::dp dp_t;
 
-  const double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
   double estimateSquaredNorm = 0;
 
   const int rowCount = block.rows->size();
@@ -453,25 +372,33 @@ compressAcaPartial(const ClusterAssemblyFunction<T>& block) {
   int J = 0;
   int k = 0;
 
+  RandomPivotManager<T> randomPivotManager(block, useRandomPivots ? max(rowCount, colCount) : 0);
+
   do {
     Vector<dp_t>* bCol = new Vector<dp_t>(block.cols->size());
     // Calculation of row I and its residue
     block.getRow(I, *bCol);
     updateRow(*bCol, I, bCols, aCols, k);
     rowFree[I] = false;
-    rowPivotCount++;
 
     // Find max and argmax of the residue
     double maxNorm2 = 0.;
     for (int j = 0; j < colCount; j++) {
-      const double norm2 = squaredNorm<dp_t>(bCol->v[j]);
+      const double norm2 = squaredNorm<dp_t>((*bCol)[j]);
       if (colFree[j] && norm2 > maxNorm2) {
         maxNorm2 = norm2;
         J = j;
       }
     }
 
-    if (bCol->v[J] == Constants<dp_t>::zero) {
+    Pivot<dp_t > randomOrDefaultPivot = randomPivotManager.GetPivot();
+    if(I!=randomOrDefaultPivot.row_ && squaredNorm(randomOrDefaultPivot.value_) > maxNorm2){
+      I = randomOrDefaultPivot.row_;
+      delete bCol;
+      continue;
+    }
+
+    if ((*bCol)[J] == Constants<dp_t>::zero) {
       delete bCol;
       // We look for another row which has not already been used.
       I = 0;
@@ -480,7 +407,7 @@ compressAcaPartial(const ClusterAssemblyFunction<T>& block) {
       }
     } else {
       // Find pivot and scale column B
-      dp_t pivot = Constants<dp_t>::pone / bCol->v[J];
+      dp_t pivot = Constants<dp_t>::pone / (*bCol)[J];
       bCol->scale(pivot);
       bCols.push_back(bCol);
 
@@ -488,13 +415,14 @@ compressAcaPartial(const ClusterAssemblyFunction<T>& block) {
       Vector<dp_t>* aCol = new Vector<dp_t>(block.rows->size());
       block.getCol(J, *aCol);
       updateCol(*aCol, J, aCols, bCols, k);
+      randomPivotManager.AddUsedPivot(bCol, aCol, I, J);
       colFree[J] = false;
       aCols.push_back(aCol);
 
       // Find max and argmax of the residue
       maxNorm2 = 0.;
       for (int i = 0; i < rowCount; i++) {
-        const double norm2 = squaredNorm<dp_t>(aCol->v[i]);
+        const double norm2 = squaredNorm<dp_t>((*aCol)[i]);
         if (rowFree[i] && norm2 > maxNorm2) {
           maxNorm2 = norm2;
           I = i;
@@ -517,42 +445,65 @@ compressAcaPartial(const ClusterAssemblyFunction<T>& block) {
       k++;
 
       // Evaluate the stopping criterion
-      // ||a_nu|| ||b_nu|| < epsilon * ||S_nu||
-      // <=> ||a_nu||^2 ||b_nu||^2 < epsilon^2 ||S_nu||^2
-      if (ab_norm_2 < epsilon * epsilon * estimateSquaredNorm) {
+      // ||a_nu|| ||b_nu|| < compressionEpsilon * ||S_nu||
+      // <=> ||a_nu||^2 ||b_nu||^2 < compressionEpsilon^2 ||S_nu||^2
+      if (ab_norm_2 < compressionEpsilon * compressionEpsilon * estimateSquaredNorm) {
         break;
       }
     }
+    rowPivotCount++;
   } while (rowPivotCount < maxK);
 
-  FullMatrix<dp_t> *newA, *newB;
+  ScalarArray<dp_t> *newA, *newB;
   if (k != 0) {
-    newA = new FullMatrix<dp_t>(block.rows->size(), k);
+    newA = new ScalarArray<dp_t>(block.rows->size(), k);
     for (int i = 0; i < k; i++) {
-      memcpy(newA->m + (i * newA->rows), aCols[i]->v, sizeof(dp_t) * newA->rows);
+      newA->copyMatrixAtOffset(aCols[i], 0, i);
       delete aCols[i];
       aCols[i] = NULL;
     }
-    newB = new FullMatrix<dp_t>(block.cols->size(), k);
+    newB = new ScalarArray<dp_t>(block.cols->size(), k);
     for (int i = 0; i < k; i++) {
-      memcpy(newB->m + (i * newB->rows), bCols[i]->v, sizeof(dp_t) * newB->rows);
+      newB->copyMatrixAtOffset(bCols[i], 0, i);
       delete bCols[i];
       bCols[i] = NULL;
     }
   } else {
     // If k == 0, block is only made of zeros.
-    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols, AcaPartial);
+    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols);
   }
 
-  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols, AcaPartial);
+  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols);
+}
+
+RkMatrix<Types<S_t>::dp>*
+CompressionAcaPartial::compress(const ClusterAssemblyFunction<S_t>& block) const {
+    return doCompressionAcaPartial<S_t>(block, epsilon_, useRandomPivots_);
+}
+RkMatrix<Types<D_t>::dp>*
+CompressionAcaPartial::compress(const ClusterAssemblyFunction<D_t>& block) const {
+    return doCompressionAcaPartial<D_t>(block, epsilon_, useRandomPivots_);
+}
+RkMatrix<Types<C_t>::dp>*
+CompressionAcaPartial::compress(const ClusterAssemblyFunction<C_t>& block) const {
+    return doCompressionAcaPartial<C_t>(block, epsilon_, useRandomPivots_);
+}
+RkMatrix<Types<Z_t>::dp>*
+CompressionAcaPartial::compress(const ClusterAssemblyFunction<Z_t>& block) const {
+    return doCompressionAcaPartial<Z_t>(block, epsilon_, useRandomPivots_);
 }
 
 
 template<typename T>
-static RkMatrix<typename Types<T>::dp>*
-compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
+RkMatrix<typename Types<T>::dp>*
+doCompressionAcaPlus(const ClusterAssemblyFunction<T>& block, double compressionEpsilon, const CompressionAlgorithm* delegate) {
+
+  if(block.cols->size() * 100 < block.rows->size() && !block.info.is_guaranteed_null_row && !block.info.is_guaranteed_null_col)
+     // ACA+ start with a findMinRow call which will last for hours
+     // if the block contains many null rows
+     return delegate->compress(block);
+
   typedef typename Types<T>::dp dp_t;
-  const double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
   double estimateSquaredNorm = 0;
   int i_ref, j_ref;
   int rowCount = block.rows->size(), colCount = block.cols->size();
@@ -564,7 +515,7 @@ compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
   j_ref = findCol(block, colFree, aRef);
   if (j_ref == -1) {
 	// The block is completely zero.
-    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols, AcaPlus);
+    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols);
   }
 
   // The reference row is chosen such that it intersects the reference
@@ -579,10 +530,10 @@ compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
     dp_t i_star_value, j_star_value;
 
     i_star = aRef.absoluteMaxIndex();
-    i_star_value = aRef.v[i_star];
+    i_star_value = aRef[i_star];
 
     j_star = bRef.absoluteMaxIndex();
-    j_star_value = bRef.v[j_star];
+    j_star_value = bRef[j_star];
 
     if (squaredNorm<dp_t>(i_star_value) > squaredNorm<dp_t>(j_star_value)) {
       // i_star is fixed, we look for j_star
@@ -590,23 +541,21 @@ compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
       // Calculate the residue
       updateRow<dp_t>(*bVec, i_star, bCols, aCols, k);
       j_star = bVec->absoluteMaxIndex();
-      dp_t pivot = bVec->v[j_star];
-      HMAT_ASSERT(pivot != Constants<dp_t>::zero);
+      dp_t pivot = (*bVec)[j_star];
       // Calculate a
       block.getCol(j_star, *aVec);
       updateCol<dp_t>(*aVec, j_star, aCols, bCols, k);
-      aVec->scale(Constants<dp_t>::pone / pivot);
+      if(pivot != Constants<dp_t>::zero) aVec->scale(Constants<dp_t>::pone / pivot);
     } else {
       // j_star is fixed, we look for i_star
       block.getCol(j_star, *aVec);
       updateCol<dp_t>(*aVec, j_star, aCols, bCols, k);
       i_star = aVec->absoluteMaxIndex();
-      dp_t pivot = aVec->v[i_star];
-      HMAT_ASSERT(pivot != Constants<dp_t>::zero);
+      dp_t pivot = (*aVec)[i_star];
       // Calculate b
       block.getRow(i_star, *bVec);
       updateRow<dp_t>(*bVec, i_star, bCols, aCols, k);
-      bVec->scale(Constants<dp_t>::pone / pivot);
+      if(pivot != Constants<dp_t>::zero) bVec->scale(Constants<dp_t>::pone / pivot);
     }
 
     rowFree[i_star] = false;
@@ -630,17 +579,17 @@ compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
     k++;
 
     // Evaluate the stopping criterion
-    // ||a_nu|| ||b_nu|| < epsilon * ||S_nu||
-    // <=> ||a_nu||^2 ||b_nu||^2 < epsilon^2 ||S_nu||^2
-    if (ab_norm_2 < epsilon * epsilon * estimateSquaredNorm) {
+    // ||a_nu|| ||b_nu|| < compressionEpsilon * ||S_nu||
+    // <=> ||a_nu||^2 ||b_nu||^2 < compressionEpsilon^2 ||S_nu||^2
+    if (ab_norm_2 < compressionEpsilon * compressionEpsilon * estimateSquaredNorm) {
       break;
     }
 
     // Update of a_ref and b_ref
-    aRef.axpy(Constants<dp_t>::mone * bCols[k - 1]->v[j_ref], aCols[k - 1]);
-    bRef.axpy(Constants<dp_t>::mone * aCols[k - 1]->v[i_ref], bCols[k - 1]);
-    const bool needNewA = isZero(aRef) || (j_star == j_ref);
-    const bool needNewB = isZero(bRef) || (i_star == i_ref);
+    aRef.axpy(Constants<dp_t>::mone * (*bCols[k - 1])[j_ref], aCols[k - 1]);
+    bRef.axpy(Constants<dp_t>::mone * (*aCols[k - 1])[i_ref], bCols[k - 1]);
+    const bool needNewA = aRef.isZero() || (j_star == j_ref);
+    const bool needNewB = bRef.isZero() || (i_star == i_ref);
 
     // If the row or the column of reference have been already chosen as pivot,
     // we can not keep it and then we take one or two others.
@@ -654,7 +603,7 @@ compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
           break;
         }
         updateCol<dp_t>(aRef, j_ref, aCols, bCols, k);
-        found = !isZero(aRef);
+        found = !aRef.isZero();
       }
       if (!found) {
         break;
@@ -683,68 +632,79 @@ compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
   } while (k < maxK);
 
   assert(k > 0);
-  FullMatrix<dp_t>* newA = new FullMatrix<dp_t>(block.rows->size(), k);
+  ScalarArray<dp_t>* newA = new ScalarArray<dp_t>(block.rows->size(), k);
   for (int i = 0; i < k; i++) {
-    memcpy(newA->m + (i * newA->rows), aCols[i]->v, sizeof(dp_t) * newA->rows);
+    newA->copyMatrixAtOffset(aCols[i], 0, i);
     delete aCols[i];
     aCols[i] = NULL;
   }
-  FullMatrix<dp_t>* newB = new FullMatrix<dp_t>(block.cols->size(), k);
+  ScalarArray<dp_t>* newB = new ScalarArray<dp_t>(block.cols->size(), k);
   for (int i = 0; i < k; i++) {
-    memcpy(newB->m + (i * newB->rows), bCols[i]->v, sizeof(dp_t) * newB->rows);
+    newB->copyMatrixAtOffset(bCols[i], 0, i);
     delete bCols[i];
     bCols[i] = NULL;
   }
-  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols, AcaPlus);
+  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols);
+}
+
+RkMatrix<Types<S_t>::dp>*
+CompressionAcaPlus::compress(const ClusterAssemblyFunction<S_t>& block) const {
+    return doCompressionAcaPlus<S_t>(block, epsilon_, delegate_);
+}
+RkMatrix<Types<D_t>::dp>*
+CompressionAcaPlus::compress(const ClusterAssemblyFunction<D_t>& block) const {
+    return doCompressionAcaPlus<D_t>(block, epsilon_, delegate_);
+}
+RkMatrix<Types<C_t>::dp>*
+CompressionAcaPlus::compress(const ClusterAssemblyFunction<C_t>& block) const {
+    return doCompressionAcaPlus<C_t>(block, epsilon_, delegate_);
+}
+RkMatrix<Types<Z_t>::dp>*
+CompressionAcaPlus::compress(const ClusterAssemblyFunction<Z_t>& block) const {
+    return doCompressionAcaPlus<Z_t>(block, epsilon_, delegate_);
 }
 
 #include <iostream>
 
 template<typename T>
-RkMatrix<typename Types<T>::dp>* compressWithoutValidation(CompressionMethod method,
-                                                           const ClusterAssemblyFunction<T>& block) {
-  typedef typename Types<T>::dp dp_t;
-  RkMatrix<dp_t>* rk = NULL;
-  switch (method) {
-  case Svd:
-    rk = compressSvd(block);
-    break;
-  case AcaFull:
-    rk = compressAcaFull(block);
-    break;
-  case AcaPartial:
-    rk = compressAcaPartial(block);
-    break;
-  case AcaPlus:
-    rk = compressAcaPlus(block);
-    break;
-  case NoCompression:
-    // Must not happen
-    HMAT_ASSERT(false);
-    break;
-  }
-
-  return rk;
+RkMatrix<typename Types<T>::dp>* compress(
+    const CompressionAlgorithm* method, const Function<T>& f,
+    const ClusterData* rows, const ClusterData* cols,
+    double epsilon, const AllocationObserver & ao) {
+    typedef typename Types<T>::dp dp_t;
+    ClusterAssemblyFunction<T> block(f, rows, cols, ao);
+    int nloop=-1; // so we assemble only one strata
+    if(block.info.number_of_strata > 1 && method->isIncremental(*rows, *cols)) {
+        block.stratum = 0;
+        // enable strata assembling for AcaPartial & AcaPlus only
+        nloop = block.info.number_of_strata;
+    }
+    RkMatrix<dp_t>* rk = compressOneStratum(method, block);
+    rk->truncate(epsilon);
+    for(block.stratum = 1; block.stratum < nloop; block.stratum++) {
+        assert(method->isIncremental(*rows, *cols));
+        RkMatrix<dp_t>* stratumRk = compressOneStratum(method, block);
+        if(stratumRk->rank() > 0) {
+            // Pass a negative value to tell formattedAddParts to not call truncate()
+            // FIXME: investigate why calling truncate from formattedAddParts or here
+            //        gives different results
+            rk->formattedAddParts(-epsilon, &Constants<dp_t>::pone, &stratumRk, 1);
+            rk->truncate(epsilon);
+        }
+        delete stratumRk;
+    }
+    return rk;
 }
 
+template<typename T> RkMatrix<typename Types<T>::dp>* compressOneStratum(
+    const CompressionAlgorithm* method, const ClusterAssemblyFunction<T> & block) {
 
-/* Appele par HMatrix<T>::assemble() */
-template<typename T>
-RkMatrix<typename Types<T>::dp>* compress(CompressionMethod method,
-                                          const Function<T>& f,
-                                          const ClusterData* rows,
-                                          const ClusterData* cols,
-                                          const AllocationObserver & ao) {
   typedef typename Types<T>::dp dp_t;
-  RkMatrix<dp_t>* rk = NULL;
-  ClusterAssemblyFunction<T> block(f, rows, cols, ao);
-
-  rk = compressWithoutValidation(method, block);
+  RkMatrix<dp_t>* rk = method->compress(block);
 
   if (HMatrix<T>::validateCompression) {
     FullMatrix<dp_t>* full = block.assemble();
-    if (rk->a) rk->a->checkNan();
-    if (rk->b) rk->b->checkNan();
+    rk->checkNan();
     FullMatrix<dp_t>* rkFull = rk->eval();
     const double approxNorm = rkFull->norm();
     const double fullNorm = full->norm();
@@ -760,25 +720,25 @@ RkMatrix<typename Types<T>::dp>* compress(CompressionMethod method,
     rkFull->axpy(Constants<T>::mone, full);
     double diffNorm = rkFull->norm();
     if (diffNorm > HMatrix<T>::validationErrorThreshold * fullNorm ) {
-      std::cout << rows->description() << "x" << cols->description() << std::endl
+      std::cout << block.rows->description() << "x" << block.cols->description() << std::endl
            << std::scientific
            << "|M|  = " << fullNorm << std::endl
            << "|Rk| = " << approxNorm << std::endl
            << "|M - Rk| / |M| = " << diffNorm / fullNorm << std::endl
-           << "Rank = " << rk->rank() << " / " << min(full->rows, full->cols) << std::endl << std::endl;
+           << "Rank = " << rk->rank() << " / " << min(full->rows(), full->cols()) << std::endl << std::endl;
 
       if (HMatrix<T>::validationReRun) {
         // Call compression a 2nd time, for debugging with gdb the work of the compression algorithm...
         RkMatrix<dp_t>* rk_bis = NULL;
 
-        rk_bis = compressWithoutValidation(method, block);
+        rk_bis = method->compress(block);
         delete rk_bis ;
       }
 
       if (HMatrix<T>::validationDump) {
         std::string filename;
         std::ostringstream convert;   // stream used for the conversion
-        convert << rows->description() << "x" << cols->description()  ;
+        convert << block.stratum << "_"<< block.rows->description() << "x" << block.cols->description();
 
         filename = "Rk_";
         filename += convert.str(); // set 'Result' to the contents of the stream
@@ -797,15 +757,15 @@ RkMatrix<typename Types<T>::dp>* compress(CompressionMethod method,
 }
 
 // Declaration of the used templates
-template RkMatrix<S_t>* compressMatrix(FullMatrix<S_t>* m, const IndexSet* rows, const IndexSet* cols);
-template RkMatrix<D_t>* compressMatrix(FullMatrix<D_t>* m, const IndexSet* rows, const IndexSet* cols);
-template RkMatrix<C_t>* compressMatrix(FullMatrix<C_t>* m, const IndexSet* rows, const IndexSet* cols);
-template RkMatrix<Z_t>* compressMatrix(FullMatrix<Z_t>* m, const IndexSet* rows, const IndexSet* cols);
+template RkMatrix<S_t>* truncatedSvd(FullMatrix<S_t>* m, double eps);
+template RkMatrix<D_t>* truncatedSvd(FullMatrix<D_t>* m, double eps);
+template RkMatrix<C_t>* truncatedSvd(FullMatrix<C_t>* m, double eps);
+template RkMatrix<Z_t>* truncatedSvd(FullMatrix<Z_t>* m, double eps);
 
-template RkMatrix<Types<S_t>::dp>* compress<S_t>(CompressionMethod method, const Function<S_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
-template RkMatrix<Types<D_t>::dp>* compress<D_t>(CompressionMethod method, const Function<D_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
-template RkMatrix<Types<C_t>::dp>* compress<C_t>(CompressionMethod method, const Function<C_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
-template RkMatrix<Types<Z_t>::dp>* compress<Z_t>(CompressionMethod method, const Function<Z_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
+template RkMatrix<Types<S_t>::dp>* compress<S_t>(const CompressionAlgorithm* method, const Function<S_t>& f, const ClusterData* rows, const ClusterData* cols, double epsilon, const AllocationObserver &);
+template RkMatrix<Types<D_t>::dp>* compress<D_t>(const CompressionAlgorithm* method, const Function<D_t>& f, const ClusterData* rows, const ClusterData* cols, double epsilon, const AllocationObserver &);
+template RkMatrix<Types<C_t>::dp>* compress<C_t>(const CompressionAlgorithm* method, const Function<C_t>& f, const ClusterData* rows, const ClusterData* cols, double epsilon, const AllocationObserver &);
+template RkMatrix<Types<Z_t>::dp>* compress<Z_t>(const CompressionAlgorithm* method, const Function<Z_t>& f, const ClusterData* rows, const ClusterData* cols, double epsilon, const AllocationObserver &);
 
 }  // end namespace hmat
 
