@@ -78,20 +78,72 @@ template<typename T> struct Types {
 template<> struct Types<S_t> {
   typedef D_t dp;
   typedef S_t sp;
+  typedef std::true_type IS_REAL;
+  static const ScalarTypes TYPE = S_TYPE;
 };
 template<> struct Types<D_t> {
   typedef D_t dp;
   typedef S_t sp;
+  typedef std::true_type IS_REAL;
+  static const ScalarTypes TYPE = D_TYPE;
 };
 template<> struct Types<C_t> {
   typedef Z_t dp;
   typedef C_t sp;
+  typedef std::false_type IS_REAL;
+  static const ScalarTypes TYPE = C_TYPE;
 };
 template<> struct Types<Z_t> {
   typedef Z_t dp;
   typedef C_t sp;
+  typedef std::false_type IS_REAL;
+  static const ScalarTypes TYPE = Z_TYPE;
 };
 
+template<typename T>
+double squaredNorm(const T x) {
+  return x * x;
+}
+
+// Specializations for complex values
+template<>
+inline double squaredNorm(const C_t x) {
+// std::norm seems deadfully slow on Intel 15
+#ifdef __INTEL_COMPILER
+  const float x_r = x.real();
+const float x_i = x.imag();
+return x_r*x_r + x_i*x_i;
+#else
+  return std::norm(x);
+#endif
+}
+
+template<>
+inline double squaredNorm(const Z_t x) {
+#ifdef __INTEL_COMPILER
+  const double x_r = x.real();
+const double x_i = x.imag();
+return x_r*x_r + x_i*x_i;
+#else
+  return std::norm(x);
+#endif
+}
+
+template<typename T>
+T conj(const T x) {
+  return x;
+}
+
+template<>
+inline C_t conj(const C_t x) {
+  return std::conj(x);
+}
+
+template<>
+inline Z_t conj(const Z_t x) {
+  return std::conj(x);
+}
+
 }  // end namespace hmat
 
 #endif
