@@ -25,39 +25,40 @@
 #include "h_matrix.hpp"
 #include "uncompressed_block.hpp"
 #include "uncompressed_values.hpp"
+#include "iengine.hpp"
 
 namespace hmat {
 
-class NullSettings {};
-template<typename T> class DefaultEngine
+template<typename T> class DefaultEngine : public IEngine<T>
 {
+  NullSettings settings;
 public:
+  ~DefaultEngine(){}
   typedef hmat::UncompressedBlock<T> UncompressedBlock;
   typedef hmat::UncompressedValues<T> UncompressedValues;
-  typedef NullSettings Settings;
-  Settings settings;
-  explicit DefaultEngine(HMatrix<T>* m = NULL): hmat(m){}
   void destroy(){}
-  // this attribute could be in HMatInterface, it's here to avoid making it friend
-  HMatrix<T>* hmat;
-  static int init() { return 0; }
+  EngineSettings& GetSettings(){ return settings;}
+  static int init();
   static void finalize(){}
-  void assembly(Assembly<T>& f, SymmetryFlag sym, bool synchronize);
-  void factorization(hmat_factorization_t);
+  void assembly(Assembly<T>& f, SymmetryFlag sym, bool ownAssembly);
+  void factorization(Factorization);
   void inverse();
-  void gemv(char trans, T alpha, FullMatrix<T>& x, T beta, FullMatrix<T>& y) const;
-  void gemm(char transA, char transB, T alpha, const DefaultEngine<T> & a, const DefaultEngine<T>& b, T beta);
+  void gemv(char trans, T alpha, ScalarArray<T>& x, T beta, ScalarArray<T>& y) const;
+  void gemm(char transA, char transB, T alpha, const IEngine<T>& a, const IEngine<T>& b, T beta);
+  void trsm(char side, char uplo, char trans, char diag, T alpha, IEngine<T> &B) const;
+  void trsm(char side, char uplo, char trans, char diag, T alpha, ScalarArray<T> &B) const;
   void addIdentity(T alpha);
-  void solve(FullMatrix<T>& b, hmat_factorization_t) const;
-  void solve(DefaultEngine<T>& b, hmat_factorization_t) const;
-  void solveLower(FullMatrix<T>& b, hmat_factorization_t t, bool transpose=false) const;
-  void copy(DefaultEngine<T> & result) const;
+  void addRand(double epsilon);
+  void solve(ScalarArray<T>& b, Factorization) const;
+  void solve(IEngine<T>& b, Factorization) const;
+  void solveLower(ScalarArray<T>& b, Factorization t, bool transpose=false) const;
+  void copy(IEngine<T> & result, bool structOnly) const;
   void transpose();
-  void createPostcriptFile(const std::string& filename) const;
-  void dumpTreeToFile(const std::string& filename, const HMatrixNodeDumper<T>& dumper_extra) const;
-  double norm() const;
-  void progress(const hmat_progress_t *){}
-  HMatrix<T> * data() const { return hmat; }
+  void applyOnLeaf(const hmat::LeafProcedure<hmat::HMatrix<T> >&f);
+  IEngine<T>* clone() const { return new DefaultEngine();};
+  HMatrix<T> * getHandle() const { return IEngine<T>::hmat; }
+  void scale(T alpha);
+  void info(hmat_info_t &i) const;
 };
 
 }  // end namespace hmat
