@@ -29,20 +29,26 @@
 
 #include "tree.hpp"
 #include "assembly.hpp"
-#include "data_types.hpp"
-#include "full_matrix.hpp"
 #include "cluster_tree.hpp"
-#include "admissibility.hpp"
+#include "rk_matrix.hpp"
+
+namespace hmat {
+    /** Identify the current user level operation */
+    enum class MainOp {OTHER, SOLVELOWER, SOLVEUPPER, GEMM};
+}
+
 #include "recursion.hpp"
 #include <cassert>
 #include <fstream>
 #include <iostream>
+#include <deque>
 
 
 namespace hmat {
 
 template<typename T> class Vector;
-template<typename T> class RkMatrix;
+class AdmissibilityCondition;
+template<typename T> class FullMatrix;
 
 /** Flag used to describe the symmetry of a matrix.
  */
@@ -50,6 +56,8 @@ enum SymmetryFlag {kNotSymmetric, kLowerSymmetric};
 
 /** Default rank value for blocks that dont have an actual computed rank
    */
+// TODO: the meaning/usage of UNINITIALIZED_BLOCK is not clear, it should be reworked
+// or removed
 enum DefaultRank {UNINITIALIZED_BLOCK = -3, NONLEAF_BLOCK = -2, FULL_BLOCK = -1};
 
 /** Settings global to a whole matrix */
@@ -59,8 +67,9 @@ struct MatrixSettings {
 /** Settings local to a matrix bloc */
 struct LocalSettings {
     const MatrixSettings * global;
-    explicit LocalSettings(const MatrixSettings * s): global(s) {}
-    //TODO add epsilons
+    LocalSettings(const MatrixSettings * s, double epsilon): global(s), epsilon_(epsilon) {}
+    /// epsilon used for SVD truncations
+    double epsilon_;
 };
 
 /** Degrees of freedom permutation of a vector required in HMatrix context.
@@ -74,7 +83,7 @@ struct LocalSettings {
      \param indices Array of indices after construction ClusterTree.
 
  */
-template<typename T> void reorderVector(FullMatrix<T>* v, int* indices);
+template<typename T> void reorderVector(ScalarArray<T>* v, int* indices, int axis);
 
 /** Inverse permutation of a vector.
 
@@ -84,36 +93,23 @@ template<typename T> void reorderVector(FullMatrix<T>* v, int* indices);
      \param indices Array of indices after construction ClusterTree.
 
  */
-template<typename T> void restoreVectorOrder(FullMatrix<T>* v, int *indices);
+template<typename T> void restoreVectorOrder(ScalarArray<T>* v, int *indices, int axis);
 
 template<typename T> class HMatrix;
-/** Class to write user defined data when dumping matrix onto disk.
 
-    This class is used by dumpTreeToFile to write extra information into
-    json file; the returned string must thus be a valid json fragment text.
+enum class Axis {ROW, COL};
+
+/** Precompute compatibility between children of a and b for GEMM.
+
+     \param a first matrix
+     \param axisA check rows or columns of a
+     \param transA tells whether a is transposed or not
+     \param b second matrix
+     \param axisB check rows or columns of b
+     \param transB tells whether b is transposed or not
+     \return byte array
  */
-template<typename T> class HMatrixNodeDumper {
-public:
-  HMatrixNodeDumper() {}
-  virtual std::string dumpExtraInfo(const HMatrix<T>& node, const std::string& prefix) const = 0;
-};
-
-template<typename T>
-class HMatrixVoidNodeDumper : public HMatrixNodeDumper<T> {
-public:
-  virtual std::string dumpExtraInfo(const HMatrix<T>&, const std::string&) const { return ""; }
-};
-
-/** Class to truncate Rk matrices.
- */
-template<typename T>
-class EpsilonTruncate : public TreeProcedure<HMatrix<T> > {
-private:
-  double epsilon_;
-public:
-  EpsilonTruncate(double epsilon) : epsilon_(epsilon) {}
-  void visit(HMatrix<T> * node, const Visit order) const;
-};
+template<typename T> unsigned char * compatibilityGridForGEMM(const HMatrix<T>* a, Axis axisA, char transA, const HMatrix<T>* b, Axis axisB, char transB);
 
 /*! \brief The HMatrix class, representing a HMatrix.
 
@@ -137,42 +133,81 @@ template<typename T> class HMatrix : public Tree<HMatrix<T> >, public RecursionM
    /// Full block, or NULL if the block is not a leaf or is compressed.
    FullMatrix<T> * full_;
   };
-  /// rank_ of the block for Rk matrices, or: UNINITIALIZED_BLOCK=-3 for an uninitialized matrix, NONLEAF_BLOCK=-2 for non leaf, FULL_BLOCK=-1 for full a matrix
+  /// rank of the block for Rk matrices, or: UNINITIALIZED_BLOCK=-3 for an uninitialized matrix, NONLEAF_BLOCK=-2 for non leaf, FULL_BLOCK=-1 for full a matrix
   int rank_;
+  /// approximate rank of the block, or: UNINITIALIZED_BLOCK=-3 for an uninitialized matrix
+  int approximateRank_;
   void uncompatibleGemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b);
   void recursiveGemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b);
   void leafGemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b);
   HMatrix<T> * fullRkSubset(const IndexSet* subset, bool col) const;
-  /*! \brief Auxiliary function used by HMatrix::dumpTreeToFile().
-   */
-  void dumpSubTree(std::ofstream& f, int depth, const HMatrixNodeDumper<T>& nodeDumper) const;
+
   /** Only used by internalCopy */
   HMatrix(const MatrixSettings * settings);
+  /** This <- This + alpha * b
+
+      \param alpha
+      \param b
+   */
+  void axpy(T alpha, const RkMatrix<T>* b);
+  /** This <- This + alpha * b
+
+      \param alpha
+      \param b
+      \param rows
+      \param cols
+   */
+  void axpy(T alpha, const FullMatrix<T>* b);
 public:
   /*! \brief Create a HMatrix based on a row and column ClusterTree.
 
     \param _rows The row cluster tree
     \param _cols The column cluster tree
    */
-  HMatrix(ClusterTree* _rows, ClusterTree* _cols, const MatrixSettings * settings,
-       SymmetryFlag symmetryFlag = kNotSymmetric,
-       AdmissibilityCondition * admissibilityCondition = &StandardAdmissibilityCondition::DEFAULT_ADMISSIBLITY);
+  HMatrix(const ClusterTree* _rows, const ClusterTree* _cols, const MatrixSettings * settings,
+       int depth, SymmetryFlag symmetryFlag,
+       AdmissibilityCondition * admissibilityCondition);
 
   /*! \brief Create a copy of this matrix for internal use only.
    * Only copy this node, not the whole tree. The created matrix
    * is an uninitialized leaf with same rows and cols as this.
    */
-  HMatrix<T> * internalCopy(bool temporary = false, bool withChildren = false) const;
+  HMatrix<T> * internalCopy(bool temporary_ = false, bool withRowChild=false, bool withColChild=false) const;
   ~HMatrix();
 
+  /** Return a full null block at the given offset and of the give size */
+  HMatrix<T> * internalCopy(const ClusterTree * rows, const ClusterTree * cols) const;
+
+  /**
+   * \brief Split this block according and admissibility condition
+   *
+   * \return true if the block was actually splitted
+  */
+  bool split(AdmissibilityCondition * admissibilityCondition, bool lowRank,
+             SymmetryFlag symmetryFlag = kNotSymmetric);
+
+  /** \brief Add the list of leaves to a list */
+  void listAllLeaves(std::deque<const HMatrix<T> *> & out) const;
+  void listAllLeaves(std::deque<HMatrix<T> *> & out);
+
+  /**
+   * Create a temporary block from a list of children.
+   * @param children The list of children ordered as insertChild and get
+   * methods expect.
+   */
+  HMatrix(const ClusterTree * rows, const ClusterTree * cols, std::vector<HMatrix*> & children);
+
   /*! \brief HMatrix coarsening.
 
      If all children are Rk leaves, then we try to merge them into a single Rk-leaf.
      This is done if the memory of the resulting leaf is less than the sum of the initial
      leaves. Note that this operation could be used hierarchically.
+     \param epsilon the truncate epsilon
      \param upper the symmetric of 'this', when building a non-sym matrix with a sym content
+     \param force if true the block is kept coarsened even if it's larger
+     \return true if all leaves are rk (i.e. if coarsening was tryed, not if it succeded)
    */
-  void coarsen(HMatrix<T>* upper = NULL) ;
+  bool coarsen(double epsilon, HMatrix<T>* upper = NULL, bool force=false) ;
   /*! \brief HMatrix assembly.
    */
   void assemble(Assembly<T>& f, const AllocationObserver & = AllocationObserver());
@@ -215,16 +250,18 @@ public:
       \param alpha scaling factor
    */
   void scale(T alpha);
-  /** Compute y <- alpha * op(this) * x + beta * y.
+  /** Compute y <- alpha * op(this) * x + beta * y if side is Side::LEFT or
+              y <- alpha * x * op(this) + beta * y if side is Side::RIGHT
 
       The arguments are similar to BLAS GEMV.
    */
-  void gemv(char trans, T alpha, const Vector<T>* x, T beta, Vector<T>* y) const;
-  /** Compute y <- alpha * op(this) * x + beta * y.
+  void gemv(char trans, T alpha, const FullMatrix<T>* x, T beta, FullMatrix<T>* y, Side side = Side::LEFT) const;
+  /** Compute y <- alpha * op(this) * x + beta * y if side is Side::LEFT or
+              y <- alpha * x * op(this) + beta * y if side is Side::RIGHT
 
       The arguments are similar to BLAS GEMV.
    */
-  void gemv(char trans, T alpha, const FullMatrix<T>* x, T beta, FullMatrix<T>* y) const;
+  void gemv(char trans, T alpha, const ScalarArray<T>* x, T beta, ScalarArray<T>* y, Side side = Side::LEFT) const;
   /*! \brief this <- alpha * op(A) * op(B) + beta * this
 
     \param transA 'N' or 'T', as in BLAS
@@ -234,7 +271,7 @@ public:
     \param b the matrix B
     \param beta beta
    */
-  void gemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b, T beta);
+  void gemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b, T beta, MainOp=MainOp::OTHER);
   /*! \brief this <- this - M * D * M^T, where 'this' is symmetric (Lower stored),
       D diagonal
 
@@ -261,45 +298,13 @@ public:
       \return a 0 matrix with the same structure as H.
    */
   static HMatrix<T>* Zero(const HMatrix<T>* h);
-  /** Create a matrix filled with 0s, based on 2 ClusterTree.
 
-      \param rows the row ClusterTree.
-      \param cols the column ClusterTree.
-      \return a 0 HMatrix.
+  /**
+   * Used internally for deserialization
+   * @see serialization.hpp
    */
-  static HMatrix<T>* Zero(const ClusterTree* rows, const ClusterTree* cols,
-                          const MatrixSettings * settings,
-                          AdmissibilityCondition * admissibilityCondition =
-                          &StandardAdmissibilityCondition::DEFAULT_ADMISSIBLITY);
-  /*! \brief Create a Postscript file representing the HMatrix.
+  static HMatrix<T> * unmarshall(const MatrixSettings * settings, int rank, int rankApprox, char bitfield, double epsilon);
 
-    The result .ps file shows the matrix structure and the compression ratio. In
-    the output, red = full block, green = compressed. The darker the green, the
-    worst the compression ration is. There is saturation at black when the block
-    size is divided by less than 5.
-
-    \param filename output filename.
-   */
-  void createPostcriptFile(const std::string& filename) const;
-  /*! \brief Dump some HMatrix metadata to a Python-readable file.
-
-    This function create a file that is readable by Python's eval()
-    function, which contains a dictionnary with the following data:
-
-    {'points': [(x1, y1, z1), ...],
-     'mapping': [indices[0], indices[1], ...],
-     'tree': {
-       'isLeaf': False,
-       'depth': 0,
-       'rows': {'offset': 0, 'n': 15243, 'boundingBox': [(-0.0249617, -0.0249652, -0.0249586), (0.0962927, 0.0249652, 0.0249688)]},
-       'cols': {'offset': 0, 'n': 15243, 'boundingBox': [(-0.0249617, -0.0249652, -0.0249586), (0.0962927, 0.0249652, 0.0249688)]},
-       'children': [child1, child2, child3, child4]
-     }
-    }
-
-    \param filename path to the output file.
-   */
-  void dumpTreeToFile(const std::string& filename, const HMatrixNodeDumper<T>& nodeDumper) const;
   /** Returns a copy of this (with all the structure and data)
        */
   HMatrix<T>* copy() const ;
@@ -322,6 +327,31 @@ public:
   double norm() const {
     return sqrt(normSqr());
   }
+
+  /*! \brief Return an approximation of the largest eigenvalue via the power method.
+   */
+  T approximateLargestEigenvalue(int max_iter, double epsilon) const;
+
+  /*! \brief Return the approximated rank.
+   */
+  int approximateRank() const {
+    return approximateRank_;
+  }
+
+  void approximateRank(int a)  {
+    approximateRank_ = a;
+  }
+
+  /*! \brief Return low rank epsilon
+   */
+  double lowRankEpsilon() const {
+    return localSettings.epsilon_;
+  }
+
+  /** Recursively set low-rank epsilon member
+   */
+  void lowRankEpsilon(double epsilon, bool recursive = true);
+
   /** Set a matrix to 0.
    */
   void clear();
@@ -337,28 +367,39 @@ public:
    */
   void transpose();
 
+  void conjugate();
   /**
    * Swap non diagonal blocks and cluster trees.
    * Only used internally.
    */
-  void transposeNoRecurse();
+  void transposeMeta(bool temporaryOnly=false);
+  /**
+   * Swap Rk or Full blocks around the diagonal
+   * Only used internally.
+   */
+  void transposeData();
 
   /*! \brief this <- o^t
 
     \param o
    */
   void copyAndTranspose(const HMatrix<T>* o);
+
+  /*! \brief Truncate Rk matrices with respect to their respective epsilon_
+   */
+  void truncate();
+
   /*! \brief LU decomposition in place.
 
     \warning Do not use. Doesn't work
    */
-  void luDecomposition();
-  /* \brief LDL^t decomposition in place
+  void luDecomposition(hmat_progress_t * progress);
+  /** \brief LDL^t decomposition in place
      \warning this has to be created with the flag lower
      \warning this has to be assembled with assembleSymmetric with onlyLower = true
    */
-  void ldltDecomposition();
-  void lltDecomposition();
+  void ldltDecomposition(hmat_progress_t * progress);
+  void lltDecomposition(hmat_progress_t * progress);
 
   /** This <- This + alpha * b
 
@@ -366,32 +407,22 @@ public:
       \param b
    */
   void axpy(T alpha, const HMatrix<T>* b);
-  /** This <- This + alpha * b
-
-      \param alpha
-      \param b
-   */
-  void axpy(T alpha, const RkMatrix<T>* b);
-  /** This <- This + alpha * b
-
-      \param alpha
-      \param b
-      \param rows
-      \param cols
-   */
-  void axpy(T alpha, const FullMatrix<T>* b, const IndexSet* rows, const IndexSet* cols);
   /** This <- This + alpha * Id
 
       \param alpha
    */
   void addIdentity(T alpha);
+  /** This <- This + A , norm(A) ~ epsilon*norm(this)
+
+      \param epsilon
+   */
+  void addRand(double epsilon);
   /*! Return true if this is a full block.
    */
   inline bool isFullMatrix() const {
     return rank_ == FULL_BLOCK && full_ != NULL;
   }
-  /* Return the full matrix corresponding to the current leaf
-   */
+  /** Return the full matrix corresponding to the current leaf */
   FullMatrix<T>* getFullMatrix() const {
     assert(isFullMatrix());
     return full_;
@@ -401,12 +432,6 @@ public:
   inline bool isRkMatrix() const {
     return rank_ >= 0;
   }
-  /*! \brief Return F * H (F Full, H divided)
-   */
-  static FullMatrix<T>* multiplyFullH(char transM, char transH, const FullMatrix<T>* m, const HMatrix<T>* h);
-  /*! \brief Return H * F (F Full, H divided)
-   */
-  static FullMatrix<T>* multiplyHFull(char transH, char transM, const HMatrix<T>* h, const FullMatrix<T>* m);
   /*! \brief Multiplication de deux HMatrix dont au moins une est une RkMatrix.
 
       Le resultat est alors une RkMatrix.
@@ -416,7 +441,7 @@ public:
       \param a A
       \param b B
    */
-  static RkMatrix<T>* multiplyRkMatrix(char transA, char transB, const HMatrix<T>* a, const HMatrix<T>* b);
+  static RkMatrix<T>* multiplyRkMatrix(double epsilon, char transA, char transB, const HMatrix<T>* a, const HMatrix<T>* b);
   /** Multiplication de deux HMatrix dont au moins une est une matrice pleine,
       et aucune n'est une RkMatrix.
 
@@ -430,76 +455,94 @@ public:
     \param left run B <- D*B instead of B <- B*D
     \param inverse run B <- B * D^-1
   */
-  void multiplyWithDiag(const HMatrix<T>* d, bool left = false, bool inverse = false) const;
+  void multiplyWithDiag(const HMatrix<T>* d, Side side = Side::RIGHT, bool inverse = false) const;
   /*! \brief Resolution du systeme L X = B, avec this = L, et X = B.
 
     \param b la matrice B en entree, et X en sortie.
    */
-  void solveLowerTriangularLeft(HMatrix<T>* b, bool unitriangular) const;
+  void solveLowerTriangularLeft(HMatrix<T>* b, Factorization algo, Diag diag, Uplo uplo, MainOp=MainOp::OTHER) const;
   /*! \brief Resolution du systeme L x = x, avec this = L, et x = b vecteur.
 
     B est un vecteur a plusieurs colonnes, donc une FullMatrix.
 
     \param b Le vecteur b en entree, et x en sortie.
    */
-  void solveLowerTriangularLeft(FullMatrix<T>* b, bool unitriangular) const;
+  void solveLowerTriangularLeft(ScalarArray<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
+  void solveLowerTriangularLeft(FullMatrix<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
   /*! Resolution de X U = B, avec U = this, et X = B.
 
     \param b la matrice B en entree, X en sortie
    */
-  void solveUpperTriangularRight(HMatrix<T>* b, bool unitriangular, bool lowerStored) const;
+  void solveUpperTriangularRight(HMatrix<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
   /*! Resolution de U X = B, avec U = this, et X = B.
 
     \param b la matrice B en entree, X en sortie
    */
-  void solveUpperTriangularLeft(HMatrix<T>* b, bool unitriangular, bool lowerStored) const;
+  void solveUpperTriangularLeft(HMatrix<T>* b, Factorization algo, Diag diag, Uplo uplo, MainOp=MainOp::OTHER) const;
   /*! Resolution de x U = b, avec U = this, et x = b.
 
     \warning b est un vecteur ligne et non colonne.
 
     \param b Le vecteur b en entree, x en sortie.
    */
-  void solveUpperTriangularRight(FullMatrix<T>* b, bool unitriangular, bool lowerStored) const;
+  void solveUpperTriangularRight(ScalarArray<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
+  void solveUpperTriangularRight(FullMatrix<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
   /*! Resolution de U x = b, avec U = this, et x = b.
     U peut etre en fait L^T ou L est une matrice stockee inferieurement
-    en precisant lowerStored = true
+    en precisant uplo = true
 
     \param b Le vecteur b en entree, x en sortie.
     \param indice les indices portes par le vecteur
-    \param lowerStored indique le stockage de la matrice U ou L^T
+    \param uplo indique le stockage de la matrice U ou L^T
   */
-  void solveUpperTriangularLeft(FullMatrix<T>* b, bool unitriangular, bool lowerStored) const;
-  /* Solve D x = b, in place with D a diagonal matrix.
+  void solveUpperTriangularLeft(ScalarArray<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
+  void solveUpperTriangularLeft(FullMatrix<T>* b, Factorization algo, Diag diag, Uplo uplo) const;
+  /*! Solve D x = b, in place with D a diagonal matrix.
 
      \param b Input: B, Output: X
    */
+  void solveDiagonal(ScalarArray<T>* b) const;
   void solveDiagonal(FullMatrix<T>* b) const;
-  void solveDiagonal(HMatrix<T>* b) const;
   /*! Resolution de This * x = b.
 
     \warning This doit etre factorisee avec \a HMatrix::luDecomposition() avant.
    */
+  void solve(ScalarArray<T>* b) const;
   void solve(FullMatrix<T>* b) const;
   /*! Resolution de This * X = b.
 
     \warning This doit etre factorisee avec \a HMatrix::luDecomposition() avant.
    */
-  void solve(HMatrix<T>* b, hmat_factorization_t) const;
+  void solve(HMatrix<T>* b, Factorization algo) const;
+
+  void trsm( char side, char uplo, char trans, char diag, T alpha,
+	     HMatrix<T>* b ) const;
+  void trsm( char side, char uplo, char trans, char diag, T alpha,
+	     ScalarArray<T>* b ) const;
+
   /*! Resolution de This * x = b.
 
     \warning This doit etre factorisee avec \a HMatrix::ldltDecomposition() avant.
    */
+  void solveLdlt(ScalarArray<T>* b) const ;
   void solveLdlt(FullMatrix<T>* b) const ;
   /*! Resolution de This * x = b.
 
     \warning This doit etre factorisee avec \a HMatrix::lltDecomposition() avant.
    */
+  void solveLlt(ScalarArray<T>* b) const ;
   void solveLlt(FullMatrix<T>* b) const ;
-  /*! Triggers an assertion is the HMatrix contains any NaN.
+  /*! Triggers an assertion if the HMatrix contains any NaN.
    */
   void checkNan() const;
-  /** Recursively set the isTriLower flag on this matrix */
+  /*! Triggers an assertion if children of an HMatrix are not contained within
+      this HMatrix.
+   */
+  void checkStructure() const;
+  /** Recursively set the isTriLower flag on this matrix diagonal blocks */
   void setTriLower(bool value);
+  /** Recursively set the isLower flag on this matrix diagonal blocks */
+  void setLower(bool value);
 
   const ClusterData* rows() const;
   const ClusterData* cols() const;
@@ -507,13 +550,15 @@ public:
   /*! \brief Return the number of children in the row dimension.
     */
   inline int nrChildRow() const {
-    return rows_->nrChild();
+    // if rows admissible, only one child = itself
+    return keepSameRows ? 1 : rows_->nrChild() ;
   }
 
   /*! \brief Return the number of children in the column dimension.
     */
   inline int nrChildCol() const {
-    return cols_->nrChild();
+    // if cols admissible, only one child = itself
+    return keepSameCols ? 1 : cols_->nrChild() ;
   }
   /*! \brief Destroy the HMatrix.
     */
@@ -532,12 +577,13 @@ public:
   HMatrix<T>* get(int i, int j) const {
     assert(i>=0 && i<nrChildRow());
     assert(j>=0 && j<nrChildCol());
+    assert(i + j * nrChildRow() < this->nrChild());
     return this->getChild(i + j * nrChildRow());
   }
 
   /*! Set the child (i, j) of this.
 
-    \warning do not use on a leaf !
+    \warning do not use on a leaf ! <- how can I know, since I am inserting a child ???
 
     \param i row index
     \param j column index
@@ -548,6 +594,18 @@ public:
     insertChild(i+j*nrChildRow(), child) ;
   }
 
+  /*! \brief Find the correct child when recursing in GEMM or GEMV
+
+    This function returns the child (i,j) of op(this) where 'op' is 'T' or 'N'.
+    If the matrix is symmetric (upper or lower), and if the child required is in the
+    part of the matrix that is not stored, it returns the symmetric block and switches 'op'.
+   \param[in,out] t input is the transpose flag for this, ouput is the transpose flag for the returned matrix
+   \param[in] i row index of the child
+   \param[in] j col index of the child
+   \return Pointer on the child
+  */
+  const HMatrix<T> * getChildForGEMM(char & t, int i, int j) const;
+
   void setClusterTrees(const ClusterTree* rows, const ClusterTree* cols);
   HMatrix<T> * subset(const IndexSet * rows, const IndexSet * cols) const;
 
@@ -558,7 +616,9 @@ public:
   /// Should try to coarsen the matrix at assembly
   static bool coarsening;
   /// Should recompress the matrix after assembly
-  static bool recompress;
+  static bool recompress;//TODO: remove
+  /// Validate the functions is_guaranteed_null_col/row() (user provided)
+  static bool validateNullRowCol;
   /// Validate the rk-matrices after compression
   static bool validateCompression;
   /// For blocks above error threshold, re-run the compression algorithm
@@ -567,10 +627,10 @@ public:
   static bool validationDump;
   /// Error threshold for the compression validation
   static double validationErrorThreshold;
-  char isUpper:1, isLower:1,       /// symmetric, upper or lower stored
+  short isUpper:1, isLower:1,       /// symmetric, upper or lower stored
        isTriUpper:1, isTriLower:1, /// upper/lower triangular
-       admissible:1, temporary:1,
-       ownClusterTree_:1;
+       keepSameRows:1, keepSameCols:1,
+       temporary_:1, ownRowsClusterTree_:1, ownColsClusterTree_:1;
   LocalSettings localSettings;
 
   int rank() const {
@@ -578,12 +638,17 @@ public:
       return rank_;
   }
 
+  /// Set the rank of an evicted rk block
+  void rank(int rank);
+
   RkMatrix<T> * rk() const {
       assert(rank_ >= 0);
       return rk_;
   }
 
-  void rk(const FullMatrix<T> * a, const FullMatrix<T> * b, bool updateRank = true);
+  /*! \brief Set 'this' as an Rk matrix using copy of a and b
+     */
+  void rk(const ScalarArray<T> *a, const ScalarArray<T> *b);
 
   void rk(RkMatrix<T> * m) {
       rk_ = m;
@@ -605,10 +670,16 @@ public:
       return rank_ == 0 || (rank_ == FULL_BLOCK && full_ == NULL);
   }
 
+  bool isRecursivelyNull() const;
+  // TODO: the meaning/usage of UNINITIALIZED_BLOCK is not clear, it should be reworked
+  // or removed
   bool isAssembled() const {
       return rank_ > UNINITIALIZED_BLOCK;
   }
 
+  bool isVoid() const {
+      return rows()->size() == 0 || cols()->size() == 0;
+  }
   /**
    * Tag a not leaf block as assembled.
    * Must only be called when all leaves of this block have been
@@ -633,6 +704,9 @@ public:
     assert(isAssembled());
   }
 
+  /** Set the entire subtree as temporary flag */
+  void temporary(bool b);
+
   const ClusterTree * rowsTree() const {
       return rows_;
   }
@@ -641,6 +715,16 @@ public:
       return cols_;
   }
 
+  void ownClusterTrees(bool owns_row, bool owns_col) {
+      ownRowsClusterTree_ = owns_row;
+      ownColsClusterTree_ = owns_col;
+  }
+
+  void setColsTree(ClusterTree * clusterTree, bool ownClusterTree) {
+      ownColsClusterTree_ = ownClusterTree;
+      cols_ = clusterTree;
+  }
+
   /**
    * Convert this HMatrix to a string for debug.
    * This is better than overriding << because it allows to use printf.
@@ -658,7 +742,6 @@ public:
       convert << "uninitialized";
     return convert.str();
   }
-
 };
 
 }  // end namespace hmat
