@@ -26,12 +26,27 @@
  */
 #ifndef HMAT_CPP_INTERFACE_HPP
 #define HMAT_CPP_INTERFACE_HPP
+
+/** C++ interface to the HMatrix library.
+
+    This is the sole entry point to the HMatrix library.
+
+    This interface is templated over the scalar type, T. This type has to be one
+    of {S_t, D_t, C_t, Z_t}, using the standard BLAS notation. For the complex
+    types, the C++ complex<> type is used. It is guaranteed to have the same
+    layout as the equivalent FORTRAN types.
+
+    The user code *has* to call \a HMatInterface<T>::init() before using any
+    other function, and \a HMatInterface<T>::finalize() at the end.
+*/
+
 #include "hmat/hmat.h"
 
 #include "clustering.hpp"
 #include "compression.hpp"
 #include "h_matrix.hpp"
-#include "default_engine.hpp"
+#include "iengine.hpp"
+#include "common/my_assert.h"
 
 namespace hmat {
 
@@ -50,33 +65,31 @@ class ClusteringAlgorithm;
 //TODO remove all global settings
 class HMatSettings: public hmat::MatrixSettings {
 public:
-  double assemblyEpsilon; ///< Tolerance for the assembly.
-  double recompressionEpsilon; ///< Tolerance for the recompression (using SVD)
-  CompressionMethod compressionMethod; ///< Compression method
   int compressionMinLeafSize; ///< Force SVD compression if max(rows->n, cols->n) < compressionMinLeafSize
+  double acaEpsilon; ///< Tolerance for the compression
+  double coarseningEpsilon; ///< Tolerance for the coarsening
   /** \f$\eta\f$ in the admissiblity condition for two clusters \f$\sigma\f$ and \f$\tau\f$:
       \f[
       \min(diam(\sigma), diam(\tau)) < \eta \cdot d(\sigma, \tau)
       \f]
    */
   int maxLeafSize; ///< Maximum size of a leaf in a ClusterTree (and of a non-admissible block in an HMatrix)
-  int maxParallelLeaves; ///< max(|L0|)
   bool coarsening; ///< Coarsen the matrix structure after assembly.
-  bool recompress; ////< Recompress the matrix after assembly.
+  bool validateNullRowCol; ///< Validate the detection of null rows and columns
   bool validateCompression; ///< Validate the rk-matrices after compression
   bool validationReRun; ///< For blocks above error threshold, re-run the compression algorithm
+  bool dumpTrace; ///< Dump trace at the end of the algorithms (depends on the runtime)
   bool validationDump; ///< For blocks above error threshold, dump the faulty block to disk
   double validationErrorThreshold; ///< Error threshold for the compression validation
 private:
   /** This constructor sets the default values.
    */
-  HMatSettings() : assemblyEpsilon(1e-4), recompressionEpsilon(1e-4),
-                   compressionMethod(AcaPlus),  compressionMinLeafSize(100),
-                   maxLeafSize(100),
-                   maxParallelLeaves(5000),
+  HMatSettings() : compressionMinLeafSize(100),
+                   coarseningEpsilon(1e-4),
+                   maxLeafSize(200),
                    coarsening(false),
-                   recompress(true), validateCompression(false),
-                   validationReRun(false), validationDump(false), validationErrorThreshold(0.) {
+                   validateNullRowCol(false), validateCompression(false),
+                   validationReRun(false), dumpTrace(false), validationDump(false), validationErrorThreshold(0.) {
     setParameters();
   }
   // Disable the copy.
@@ -95,27 +108,13 @@ private:
       This method has to be called for the settings to take effect.
    */
   void setParameters() const;
-  /** Output a textual representation of the settings to out.
-
-      @param out The output stream, std::cout by default.
-   */
-  void printSettings(std::ostream& out = std::cout) const;
 };
 
 DofCoordinates* createCoordinates(double* coord, int dim, int size);
 
-/** Create a ClusterTree.
-
-    The exact type of the returned ClusterTree depends on the global HMatrix
-    library settings (\a HMatSettings::clustering).  Passing the returned
-    ClusterTree pointer to the constructor of \a HMatInterface transfers its
-    ownership to this instance. It is then automatically freed at the destrution
-    of its owner.
-
-    @note This is the only proper way to dispose of a ClusterTree instance.
-
-    @param dls Array of DofCoordinate, of length n
-    @return a ClusterTree instance.
+/**
+ * @deprecated use one of hmat_create_cluster_tree, hmat_create_cluster_tree_from_builder,
+ * hmat_create_cluster_tree_generic
  */
 ClusterTree* createClusterTree(const DofCoordinates& dls, const ClusteringAlgorithm& algo = MedianBisectionAlgorithm());
 
@@ -130,38 +129,13 @@ class DefaultProgress
         void operator=(DefaultProgress const&);
 };
 
-/** C++ interface to the HMatrix library.
-
-    This is the sole entry point to the HMatrix library.
-
-    This interface is templated over the scalar type, T. This type has to be one
-    of {S_t, D_t, C_t, Z_t}, using the standard BLAS notation. For the complex
-    types, the C++ complex<> type is used. It is guaranteed to have the same
-    layout as the equivalent FORTRAN types.
-
-    The user code *has* to call \a HMatInterface<T>::init() before using any
-    other function, and \a HMatInterface<T>::finalize() at the end.
-*/
-template<typename T, template <typename> class E = DefaultEngine>
+    template<typename T>
 class HMatInterface {
- private:
-  static bool initialized; ///< True if the library has been initialized.
-
 private:
-  E<T> engine_;
-  hmat_factorization_t factorizationType;
+  IEngine<T>* engine_;
+  Factorization factorizationType;
 
 public:
-  /** Initialize the library.
-
-      @warning This *must* be called before using the HMatrix library.
-   */
-  static int init();
-  /** Finalize the library.
-
-      @warning The library cannot be used after this has been called.
-   */
-  static void finalize();
   /** Build a new HMatrix from two cluster sets.
 
       @note The ownership of the two ClusterTree instances (which don't need to
@@ -174,9 +148,11 @@ public:
       @return a new HMatInterface instance.
    */
 
-  HMatInterface(ClusterTree* _rows, ClusterTree* _cols, SymmetryFlag sym,
-                AdmissibilityCondition * admissibilityCondition =
-                &StandardAdmissibilityCondition::DEFAULT_ADMISSIBLITY);
+  HMatInterface(IEngine<T>* engine, const ClusterTree* _rows, const ClusterTree* _cols, SymmetryFlag sym,
+                AdmissibilityCondition * admissibilityCondition);
+
+  HMatInterface(IEngine<T>* engine, HMatrix<T>* h, Factorization factorization = Factorization::NONE);
+
   /** Destroy an HMatInterface instance.
 
       @note This destructor is *not* virtual, as this class is not meant to be subclassed.
@@ -186,16 +162,17 @@ public:
 
       This builds an HMatrix using a provided AssemblyFunction. The compression
       method is determined by \a HMatSettings::compressionMethod, and the
-      tolerance by \a HMatSettings::assemblyEpsilon. A recompression is done
-      when \a HMatSettings::recompress is true.
+      tolerance by \a HMatSettings::acaEpsilon.
 
       @param f The assembly function used to compute various matrix sub-parts
       @param sym If kLowerSymmetric, compute only the lower triangular matrix, and transpose
                  block to store upper counterpart.
-      @param synchronize
+      @param s: deprecated parameter
+      @param ownAssembly true if &f should be deleted by the assemble function
    */
-  void assemble(Assembly<T>& f, SymmetryFlag sym, bool synchronize=true,
-                hmat_progress_t * progress = DefaultProgress::getInstance());
+  void assemble(Assembly<T>& f, SymmetryFlag sym, bool s = true,
+                hmat_progress_t * progress = DefaultProgress::getInstance(),
+                bool ownAssembly=false);
 
   /** Compute a \f$LU\f$ or \f$LDL^T\f$ decomposition of the HMatrix, in place.
 
@@ -204,7 +181,7 @@ public:
       HMatInterface<T>::assemble()), and if HMatSettings::useLdlt is
       true. Otherwise an LU decomposition is done.
    */
-  void factorize(hmat_factorization_t, hmat_progress_t * progress = DefaultProgress::getInstance());
+  void factorize(Factorization, hmat_progress_t * progress = DefaultProgress::getInstance());
 
   /** Compute the inverse of the HMatrix, in place.
    */
@@ -213,7 +190,7 @@ public:
   /** Matrix-Vector product.
 
       This computes \f$ y \gets \alpha . op(A) x + \beta y\f$, with A = this, x
-      and y FullMatrix<T>. If trans == 'N', then op(A) = A, if trans == 'T',
+      and y ScalarArray<T>. If trans == 'N', then op(A) = A, if trans == 'T',
       then op(A) = A^T, as in BLAS.
 
       @param trans 'N' or 'T'
@@ -222,7 +199,8 @@ public:
       @param beta
       @param y
    */
-  void gemv(char trans, T alpha, FullMatrix<T>& x, T beta, FullMatrix<T>& y) const;
+  void gemv(char trans, T alpha, ScalarArray<T>& x, T beta, ScalarArray<T>& y) const;
+  void gemm_scalar(char trans, T alpha, ScalarArray<T>& x, T beta, ScalarArray<T>& y) const;
   /** Matrix-Matrix product.
 
       This computes \f$ C \gets \alpha . op(A) \times op(B) + \beta C\f$ with A,
@@ -238,39 +216,67 @@ public:
       @param b
       @param beta
    */
-  void gemm(char transA, char transB, T alpha, const HMatInterface<T, E>* a, const HMatInterface<T, E>* b, T beta);
+  void gemm(char transA, char transB, T alpha, const HMatInterface<T>* a, const HMatInterface<T>* b, T beta);
+  /** Matrix triangular solve.
+
+      This solves \f$ alpha op(A) X = B or X op(A) = B\f$ where B is
+      replaced by X on output.  A is a HMatInterface<T> instance, and
+      B is either an HMatInterface<T> instance, or a ScalarArray<T>.
+      A = this.
+      If transa == 'N' then op(A) = A, if transa == 'T', then op(A) =
+      A^T, and if transa == 'C', then op(A) = A^H as in BLAS.
+      If side == 'L', then \f$ op(A) X = B \f$ is solved, else if 
+      side == 'R', then \f$ X op(A) = B \f$ is solved.
+      If uplo == 'U', then A is upper triangular, else if uplo == 'L',
+      A is lower triangular.
+      If diag == 'U', then A is unit triangular, else A is non unit triangular.
+
+      @note cases LUT, RUT, and RLN are not supported.
+
+      @param side   'L' or 'R'
+      @param uplo   'U' or 'L'
+      @param transA 'N' or 'T' or 'C'
+      @param diag   'U' or 'N'
+      @param alpha
+      @param B
+   */
+  void trsm( char side, char uplo, char transa, char diag,
+	     T alpha, HMatInterface<T>* B );
+  void trsm( char side, char uplo, char transa, char diag,
+	     T alpha, ScalarArray<T>& B );
+
   /** Full <- Full x HMatrix product.
 
       This computes the product \f$ C_F \gets \alpha . op(A_F) \times op(B_H) +
-      \beta C_F\f$, with \f$A_F\f$, \f$C_F\f$ two FullMatrix<T>, and \f$B_H\f$
+      \beta C_F\f$, with \f$A_F\f$, \f$C_F\f$ two ScalarArray<T>, and \f$B_H\f$
       an HMatrixInterface<T> instance.
 
       The meaning of the arguments is as in \a HMatInterface<T>::gemm(), and in
       BLAS.
    */
-  static void gemm(FullMatrix<T>& c, char transA, char transB, T alpha, FullMatrix<T>& a, const HMatInterface<T, E>& b, T beta);
+  static void gemm(ScalarArray<T>& c, char transA, char transB, T alpha, ScalarArray<T>& a, const HMatInterface<T>& b, T beta);
   /** Return a new copy of this.
    */
-  HMatInterface<T, E>* copy() const;
+  HMatInterface<T>* copy(bool structOnly = false) const;
   /** Transpose this in place.
    */
   void transpose();
-  /** Solve the system \f$A x = b\f$ in place, with A = this, and b a FullMatrix.
+  /** Solve the system \f$A x = b\f$ in place, with A = this, and b a ScalarArray.
 
       @warning A has to be factored first with \a HMatInterface<T>::factorize().
    */
-  void solve(FullMatrix<T>& b) const;
+  void solve(ScalarArray<T>& b) const;
   /** Solve the system \f$A x = B\f$ in place, with A = this, and B a HMatInterface<T>.
 
       @warning A has to be factored first with \a HMatInterface<T>::factorize().
    */
-  void solve(HMatInterface<T, E>& b) const;
+  void solve(HMatInterface<T>& b) const;
   /** Solve the system \f$op(L) x = b\f$ in place, with L being the lower triangular part of
-      an already factorized matrix, and b a FullMatrix.
+      an already factorized matrix, and b a ScalarArray.
 
       @warning A has to be factored first with \a HMatInterface<T>::factorize().
    */
-  void solveLower(FullMatrix<T>& b, bool transpose=false) const;
+  void solveLower(ScalarArray<T>& b, bool transpose=false) const;
   /** Return an approximation of the Frobenius norm of this.
    */
   double norm() const;
@@ -278,26 +284,23 @@ public:
    */
   void scale(T alpha);
 
+  /** Recompress all Rk matrices to their respective epsilon_ values.
+   */
+  void truncate();
+
   /** this <- this + alpha * Id
    */
   void addIdentity(T alpha);
 
+  /** this <- this + A(norm = epsilon * norm(this)
+   */
+  void addRand(double epsilon);
   /**
    * Fill a hmat_info_t structure with information of this matrix.
    * @note This is only meaningful once the HMatrix has been assembled.
    */
   void info(hmat_info_t &) const;
 
-  /** Create a Postscript file representing the HMatrix.
-
-    The result .ps file shows the matrix structure and the compression ratio. In
-    the output, red = full block, green = compressed. The darker the green, the
-    worst the compression ration is. There is saturation at black when the block
-    size is divided by less than 5.
-
-    @param filename output filename.
-   */
-  void createPostcriptFile(const std::string& filename) const;
   /*! \brief Dump some HMatrix metadata to a Python-readable file.
 
     This function create a file that is readable by Python's eval()
@@ -317,7 +320,7 @@ public:
     \param filename path to the output file.
    */
   void dumpTreeToFile(const std::string& filename) const;
-  void dumpTreeToFile(const std::string& filename, const HMatrixNodeDumper<T>& dumper_extra) const;
+
   /** Return the number of block cluster tree nodes.
    */
   int nodesCount() const;
@@ -325,25 +328,44 @@ public:
    */
   void walk(TreeProcedure<HMatrix<T> > *proc);
 
-  typename E<T>::Settings & engineSettings() { return engine_.settings; }
+  /** Recursively apply a procedure to all leaves of an HMatrix.
+   */
+  void apply_on_leaf(const LeafProcedure<HMatrix<T> >& proc);
 
   const ClusterData * rows() const {
-      return engine_.hmat->rows();
+      return engine_->hmat->rows();
   }
 
   const ClusterData * cols() const {
-      return engine_.hmat->cols();
+      return engine_->hmat->cols();
   }
 
-  const E<T> & engine() const {
-      return engine_;
+  const IEngine<T> & engine() const {
+      return *engine_;
   }
 
+  EngineSettings& engineSettings() {
+    return engine_->GetSettings();
+  }
+
+  Factorization factorization() {
+      return factorizationType;
+  }
+ 
+  HMatrix<T>* get( int i, int j) const;
+
+  void setHMatrix( HMatrix<T> *hmat = NULL) const {
+      engine_->setHMatrix( hmat );
+  }
+
+  void progress(hmat_progress_t * progress) {
+      engine_->progress(progress);
+  }
 private:
-  HMatInterface(HMatrix<T>* h);
   /// Disallow the copy
-  HMatInterface(const HMatInterface<T, E>& o);
+  HMatInterface(const HMatInterface<T>& o);
 };
+
 }  // end namespace hmat
 
 #endif
