@@ -324,9 +324,9 @@ inline int
 gesvd<hmat::C_t, hmat::S_t>(char jobu, char jobvt, int m, int n, hmat::C_t* a, int lda,  hmat::S_t* s, hmat::C_t* u, int ldu, hmat::C_t* vt, int ldvt, hmat::C_t* work, int lwork) {
   int info = 0;
   const int mn = (m < n ? m : n);
-  hmat::S_t* rwork = (lwork == -1 ? NULL : new hmat::S_t[5 * mn]);
+  hmat::S_t* rwork = new hmat::S_t[ mn == 0 ? 1 : 5*mn ];
   _CGESVD_(&jobu, &jobvt, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, rwork, &info);
-  if (rwork) delete [] rwork;
+  delete [] rwork;
   return info;
 }
 template<>
@@ -334,9 +334,12 @@ inline int
 gesvd<hmat::Z_t, hmat::D_t>(char jobu, char jobvt, int m, int n, hmat::Z_t* a, int lda,  hmat::D_t* s, hmat::Z_t* u, int ldu, hmat::Z_t* vt, int ldvt, hmat::Z_t* work, int lwork) {
   int info = 0;
   const int mn = (m < n ? m : n);
-  hmat::D_t* rwork = (lwork == -1 ? NULL : new hmat::D_t[5 * mn]);
+  // For rwork, MKL says: Workspace array, size at least max(1, 5*min(m, n)). Used in complex flavors only.
+  //            NETLIB says: (workspace) DOUBLE PRECISION array, dimension (5*min(M,N))
+  // We do like MKL says to please both
+  hmat::D_t* rwork = new hmat::D_t[ mn == 0 ? 1 : 5*mn ];
   _ZGESVD_(&jobu, &jobvt, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, rwork, &info);
-  if (rwork) delete [] rwork;
+  delete [] rwork;
   return info;
 }
 #undef _SGESVD_
@@ -565,6 +568,19 @@ template <> inline int potrf<hmat::Z_t>(char uplo, int n, hmat::Z_t* a, int lda)
 #undef _CPOTRF_
 #undef _ZPOTRF_
 
+#define _CLACGV_ F77_FUNC(clacgv,CLACGV)
+#define _ZLACGV_ F77_FUNC(zlacgv,ZLACGV)
+extern "C" {
+void _CLACGV_(int* n, hmat::C_t* a, int* incx);
+void _ZLACGV_(int* n, hmat::Z_t* a, int* incx);
+}
+template<typename T> void lacgv(int n, T* a, int incx);
+template<> inline void lacgv<hmat::S_t>(int, hmat::S_t*, int) {}
+template<> inline void lacgv<hmat::D_t>(int, hmat::D_t*, int) {}
+template<> inline void lacgv<hmat::C_t>(int n, hmat::C_t* a, int incx) { _CLACGV_(&n, a, &incx); }
+template<> inline void lacgv<hmat::Z_t>(int n, hmat::Z_t* a, int incx) { _ZLACGV_(&n, a, &incx); }
+#undef _CLACGV_
+#undef _ZLACGV_
 
 }  // end namespace proxy_lapack
 
