@@ -29,11 +29,12 @@
 
 #include "full_matrix.hpp"
 #include "compression.hpp"
+#include "common/my_assert.h"
 
 namespace hmat {
 
 template<typename T> class HMatrix;
-template<typename T> class AssemblyFunction;
+template<typename T, template <typename> class F> class AssemblyFunction;
 class ClusterData;
 class IndexSet;
 
@@ -46,35 +47,19 @@ class IndexSet;
  */
 class RkApproximationControl {
 public:
-  int k; /// If != 0, fixed-rank approximation
-  double assemblyEpsilon; /// Tolerance for the assembly
-  double recompressionEpsilon; /// Tolerance for the recompressions
-  CompressionMethod method;
+  double coarseningEpsilon; /// Tolerance for the coarsening
   int compressionMinLeafSize;
 
   /** Initialization with impossible values by default
    */
-  RkApproximationControl() : k(0), assemblyEpsilon(-1.),
-                             recompressionEpsilon(-1.), method(Svd), compressionMinLeafSize(100) {}
-  /** Returns the number of singular values to keep.
-
-       The stop criterion is (assuming that the singular value
-       are in descending order):
-           sigma [k] / SUM (sigma) <epsilon
-
-       \param sigma table of singular values at least maxK elements.
-       \param maxK maximum number of singular values to keep.
-       \param epsilon tolerance.
-       \return int the number of singular values to keep.
-
-       note : the parameters maxK and sigma seem have contradictory explanation
-   */
-  int findK(double *sigma, int maxK, double epsilon);
+  RkApproximationControl() : coarseningEpsilon(-1.),
+                             compressionMinLeafSize(100) {}
 };
 
 
 template<typename T> class RkMatrix {
-
+  // Type double precision associated to T
+  typedef typename Types<T>::dp dp_t;
   /**  Swaps members of two RkMatrix instances.
        Since rows and cols are constant, they cannot be swaped and
        the other instance must have the same members.
@@ -82,23 +67,26 @@ template<typename T> class RkMatrix {
        \param other  other RkMatrix instance.
   */
   void swap(RkMatrix<T>& other);
+  /** Recompress an RkMatrix in place with a modified Gram-Schmidt algorithm.
 
+      @warning The previous rk->a and rk->b are no longer valid after this function.
+      \param epsilon is the accuracy of the recompression
+      \param initialPivotA/B is the number of orthogonal columns in panels a and b
+   */
+  void mGSTruncate(double epsilon, int initialPivotA=0, int initialPivotB=0);
 public:
+  /** @brief A hook which can be called at the begining of formatedAddParts */
+  static bool (*formatedAddPartsHook)(RkMatrix<T> * me, double epsilon, const T* alpha, const RkMatrix<T>* const * parts, const int n);
   const IndexSet *rows;
   const IndexSet *cols;
   // A B^t
-  FullMatrix<T>* a;
-  FullMatrix<T>* b;
-  CompressionMethod method; /// Method used to compress this RkMatrix
+  ScalarArray<T>* a;
+  ScalarArray<T>* b;
 
-public:
   /// Control of the approximation. See \a RkApproximationControl for more
   /// details.
   static RkApproximationControl approx;
 
-  // Type double precision associated to T
-  typedef typename Types<T>::dp dp_t;
-
   /** Construction of a RkMatrix .
 
        A Rk-matrix is a compressed representation of a matrix of size
@@ -110,16 +98,15 @@ public:
        \param _b matrix B (not B^t)
        \param _cols indices of the columns (of size k)
    */
-  RkMatrix(FullMatrix<T>* _a, const IndexSet* _rows,
-           FullMatrix<T>* _b, const IndexSet* _cols,
-           CompressionMethod _method);
+  RkMatrix(ScalarArray<T>* _a, const IndexSet* _rows,
+           ScalarArray<T>* _b, const IndexSet* _cols);
   ~RkMatrix();
 
   int rank() const {
       return a ? a->cols : 0;
   }
 
-  /**  Gives a pointer to a RkMatrix representing a subset of indices.
+  /**  Returns a pointer to a new RkMatrix representing a subset of indices.
        The pointer is supposed to be read-only (for efficiency reasons).
 
        \param subRows subset of rows
@@ -127,20 +114,27 @@ public:
        \return pointer to a new matrix with subRows and subCols.
    */
   const RkMatrix* subset(const IndexSet* subRows, const IndexSet* subCols) const;
+  RkMatrix* truncatedSubset(const IndexSet* subRows, const IndexSet* subCols, double epsilon) const;
   /** Returns the compression ratio (stored_elements, total_elements).
    */
   size_t compressedSize();
 
   size_t uncompressedSize();
 
-  /** Returns a pointer to a new matrix M = AB^t (uncompressed)
+  /** Returns a pointer to a new FullMatrix M = AB^t (uncompressed)
    */
   FullMatrix<T>* eval() const;
+  /** Returns a pointer to a new ScalarArray M = AB^t (uncompressed) or fill an existing one
+   */
+  ScalarArray<T>* evalArray(ScalarArray<T> *result=NULL) const ;
   /** Recompress an RkMatrix in place.
 
       @warning The previous rk->a and rk->b are no longer valid after this function.
+      \param initialPivotA/B is the number of orthogonal columns in panels a and b
    */
-  void truncate(double epsilon);
+  void truncate(double epsilon, int initialPivotA=0, int initialPivotB=0);
+  /** Add randomness to the RkMatrix */
+  void addRand(double epsilon);
   /*! \brief Return square of the Frobenius norm of the matrix.
 
     \return the matrix norm.
@@ -151,43 +145,13 @@ public:
       \param alpha
       \param mat
    */
-  void axpy(T alpha, const FullMatrix<T>* mat);
+  void axpy(double epsilon, T alpha, const FullMatrix<T>* mat);
   /** this <- this + alpha * mat
 
       \param alpha
       \param mat
    */
-  void axpy(T alpha, const RkMatrix<T>* mat);
-  /** Formatted addition of two Rk-matrices.
-
-      The two matrices must be on the same sets of indices in the case
-      otherwise use RkMatrix::formattedAddParts. The formatted addition of R
-      and S is defined by:
-       \code
-       truncate (R + S)
-       \endcode
-      with the addition defined by the juxtaposition of the matrices A and B of
-      each RkMatrix component of the product.
-
-      \param o The matrix sum
-      \return truncate(*this + m) A new matrix.
-   */
-  RkMatrix<T> *formattedAdd(const FullMatrix<T>* o, T alpha = Constants<T>::pone) const;
-  /** Addition of a matrix and formatted a RkMatrix.
-
-      The two matrices must be on the same sets of indices in the case
-      otherwise use RkMatrix::formattedAddParts. The formatted addition of R
-      and S is defined by:
-       \code
-       truncate(R + S)
-       \endcode
-      with the addition defined by the juxtaposition of the matrices A and B of
-      RkMatrix each component of the product.
-
-      \param o The matrix sum
-      \return truncate(*this + m) A new matrix.
-   */
-  RkMatrix<T> *formattedAdd(const RkMatrix<T>* o) const;
+  void axpy(double epsilon, T alpha, const RkMatrix<T>* mat);
   /** Adds a list of RkMatrix to a RkMatrix.
 
       In this function, RkMatrix may include some
@@ -196,9 +160,12 @@ public:
 
       \param units The list RkMatrix adding.
       \param n Number of matrices to add
+      \param epsilon truncate epsilon (negative value disable truncate)
+      \param hook true to call formattedAddPartsHook, false to do not call
       \return truncate(*this + parts[0] + parts[1] + ... + parts[n-1])
    */
-  RkMatrix<T>* formattedAddParts(T* alpha, const RkMatrix<T>** parts, int n) const;
+  void formattedAddParts(double epsilon, const T* alpha, const RkMatrix<T>* const * parts, const int n,
+                                 bool hook = true);
   /** Adds a list of MatrixXd (solid matrices) to RkMatrix.
 
       In this function, MatrixXd may cover a portion of
@@ -208,31 +175,40 @@ public:
       with RkMatrix :: formattedAddParts.
 
       \param units The list MatrixXd adding.
-      \param rowsList The list of indices rows
-      \param colsList The list of column indices
       \param n Number of matrices to add
       \return truncate (*this + parts[0] + parts[1] + ... + parts[n-1])
    */
-  RkMatrix<T>* formattedAddParts(T* alpha, const FullMatrix<T>** parts,
-                                 const IndexSet** rowsList,
-                                 const IndexSet** colsList, int n) const;
-  void gemmRk(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>* b, T beta);
+  void formattedAddParts(double epsilon, const T* alpha, const FullMatrix<T>* const * parts, int n);
+
+  /*! \brief Add a product of HMatrix to an RkMatrix
+     */
+  void gemmRk(double epsilon, char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>* b);
 
   /** Multiplication by a scalar.
 
        \param alpha the scalar
    */
   void scale(T alpha);
+  /** \brief Transpose in place.
+   */
+  void transpose();
   void clear();
+
   /** Copy  RkMatrix into this.
    */
-  void copy(RkMatrix<T>* o);
+  void copy(const RkMatrix<T>* o);
 
-  /** Compute y <- alpha * op(A) * y + beta * y.
+  /** Return a copy of this.
+   */
+  RkMatrix<T>* copy() const;
+
+  /** Compute y <- alpha * op(A) * y + beta * y if side is Side::LEFT
+           or y <- alpha * y * op(A) + beta * y if side is Side::RIGHT
+      with x and y ScalarArray<T>*.
 
       The arguments are similar to BLAS GEMV.
    */
-  void gemv(char trans, T alpha, const FullMatrix<T>* x, T beta, FullMatrix<T>* y) const;
+  void gemv(char trans, T alpha, const ScalarArray<T>* x, T beta, ScalarArray<T>* y, Side side = Side::LEFT) const;
 
   /**  Right multiplication of RkMatrix by a matrix.
 
@@ -241,8 +217,7 @@ public:
        \return R * M
   */
   static RkMatrix<T>* multiplyRkFull(char transR, char transM,
-                                     const RkMatrix<T>* rk, const FullMatrix<T>* m,
-                                     const IndexSet* mCols);
+                                     const RkMatrix<T>* rk, const FullMatrix<T>* m);
 
   /** Left multiplication of RkMatrix by a matrix.
 
@@ -253,8 +228,7 @@ public:
    */
   static RkMatrix<T>* multiplyFullRk(char transM, char transR,
                                      const FullMatrix<T>* m,
-                                     const RkMatrix<T>* rk,
-                                     const IndexSet* mRows);
+                                     const RkMatrix<T>* rk);
   /* These functions are added to manage the particular case of the product by
       H-matrix, which is treated by decomposing the product into the succession of
       products by a vector, the result being a RkMatrix.
@@ -289,20 +263,33 @@ public:
        \param b
        \return A * B
   */
-  static RkMatrix<T>* multiplyRkRk(char transA, char transB, const RkMatrix<T>* a, const RkMatrix<T>* b);
+  static RkMatrix<T>* multiplyRkRk(char transA, char transB, const RkMatrix<T>* a, const RkMatrix<T>* b, double epsilon);
   /*! \brief in situ multiplication of the matrix by the diagonal of the matrix given as argument
 
      \param d D matrix which we just considered the diagonal
      \param inverse D or D^{-1}
-     \param left multiplication (left = true) or right (left = false) of this by D
+     \param left multiplication (side = Side::LEFT) or right (side = Side::RIGHT) of this by D
   */
-  void multiplyWithDiagOrDiagInv(const HMatrix<T> * d, bool inverse, bool left = false);
+  void multiplyWithDiagOrDiagInv(const HMatrix<T> * d, bool inverse, Side side = Side::RIGHT);
   /*! \brief Triggers an assertion if there are NaNs in the RkMatrix.
    */
   void checkNan() const;
 
+  /*! Conjugate the content of the complex matrix */
+  void conjugate();
+
   /** Return the memory size of the a*b product */
   static size_t computeRkRkMemorySize(char transA, char transB, const RkMatrix<T>* a, const RkMatrix<T>* b);
+
+  /** Simpler accessor for the data.
+
+      There is only 1 type to because we cant modify the data.
+   */
+  T get(int i, int j) const ;
+
+  /*! \brief Write the RkMatrix data 'a' and 'b' in a stream (FILE*, unix fd, ...)
+    */
+  void writeArray(hmat_iostream writeFunc, void * userData) const;
 };
 
 }  // end namespace hmat
